import ee
import pandas as pd
from pathlib import Path
from .datasets import combine_datasets
import json
import logging
import country_converter as coco
from openforis_whisp.parameters.config_runtime import (
    plot_id_column,
    external_id_column,
    geometry_type_column,
    geometry_area_column,
    geometry_area_column_formatting,
    centroid_x_coord_column,
    centroid_y_coord_column,
    iso3_country_column,
    iso2_country_column,
    admin_1_column,
    stats_unit_type_column,
    stats_area_columns_formatting,
    stats_percent_columns_formatting,
    water_flag,
)
from .data_conversion import (
    convert_ee_to_df,
    convert_geojson_to_ee,
    convert_ee_to_geojson,
    # convert_csv_to_geojson,
    convert_df_to_geojson,
)  # copied functions from whisp-api and geemap (accessed 2024) to avoid dependency
from .reformat import (
    validate_dataframe_using_lookups,
    validate_dataframe_using_lookups_flexible,
)

# NB functions that included "formatted" in the name apply a schema for validation and reformatting of the output dataframe. The schema is created from lookup tables.

# ============================================================================
# PERFORMANCE OPTIMIZATION: Cache expensive Earth Engine datasets
# ============================================================================
# These images/collections are loaded once and reused across all features
# to avoid repeated expensive operations. This saves 7-15 seconds per analysis.

_WATER_FLAG_IMAGE = None
_admin_boundaries_FC = None


def get_water_flag_image():
    """
    Get cached water flag image.

    OPTIMIZATION: Water flag image is created once and reused for all features.
    This avoids recreating ocean/water datasets for every feature (previously
    called in get_type_and_location for each feature).

    Returns
    -------
    ee.Image
        Cached water flag image
    """
    global _WATER_FLAG_IMAGE
    if _WATER_FLAG_IMAGE is None:
        _WATER_FLAG_IMAGE = water_flag_all_prep()
    return _WATER_FLAG_IMAGE


def get_admin_boundaries_fc():
    """
    Get cached GAUL 2024 L1 administrative boundary feature collection.

    OPTIMIZATION: GAUL 2024 L1 collection is loaded once and reused for all features.
    This avoids loading the large FeatureCollection for every feature (previously
    called in get_admin_boundaries_info for each feature).

    Returns
    -------
    ee.FeatureCollection
        Cached GAUL 2024 L1 administrative boundary feature collection
    """
    global _admin_boundaries_FC
    if _admin_boundaries_FC is None:
        _admin_boundaries_FC = ee.FeatureCollection(
            "projects/sat-io/open-datasets/FAO/GAUL/GAUL_2024_L1"
        )
    return _admin_boundaries_FC


def whisp_formatted_stats_geojson_to_df_legacy(
    input_geojson_filepath: Path | str,
    external_id_column=None,
    national_codes=None,
    unit_type="ha",
    whisp_image=None,
    custom_bands=None,  # New parameter
) -> pd.DataFrame:
    """
        Legacy function for basic Whisp stats extraction.

        DEPRECATED: This is the original implementation maintained for backward compatibility.
        Use whisp_formatted_stats_geojson_to_df() for new code, which provides automatic
        optimization, formatting, and schema validation.

        Converts a GeoJSON file to a pandas DataFrame containing Whisp stats for the input ROI.
        Output df is validated against a panderas schema (created on the fly from the two lookup CSVs).

        This function first converts the provided GeoJSON file into an Earth Engine FeatureCollection.
        It then processes the FeatureCollection to extract relevant Whisp statistics,
        returning a structured DataFrame that aligns with the expected schema.

        If `external_id_column` is provided, it will be used to link external identifiers
        from the input GeoJSON to the output DataFrame.

        Parameters
        ----------
        input_geojson_filepath : Path | str
            The filepath to the GeoJSON of the ROI to analyze.
        external_id_column : str, optional
            The column in the GeoJSON containing external IDs to be preserved in the output DataFrame.
            This column must exist as a property in ALL features of the GeoJSON file.
            Use debug_feature_collection_properties() to inspect available properties if you encounter errors.
        remove_geom : bool, default=False
            If True, the geometry of the GeoJSON is removed from the output DataFrame.
        national_codes : list, optional
            List of ISO2 country codes to include national datasets.
        unit_type: str, optional
            Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
        whisp_image : ee.Image, optional
            Pre-combined multiband Earth Engine Image containing all Whisp datasets.
            If provided, this image will be used instead of combining datasets based on national_codes.
            If None, datasets will be combined automatically using national_codes parameter.
        custom_bands : list or dict, optional
            Custom band information for extra columns. Can be:
            - List of band names: ['Aa_test', 'elevation']
            - Dict with types: {'Aa_test': 'float64', 'elevation': 'float32'}
            - None: preserves all extra columns automatically

    Returns
        -------
        df_stats : pd.DataFrame
            The DataFrame containing the Whisp stats for the input ROI.
    """
    # Convert GeoJSON to Earth Engine FeatureCollection
    # Note: Geometry validation/cleaning should be done before calling this function
    feature_collection = convert_geojson_to_ee(str(input_geojson_filepath))

    return whisp_formatted_stats_ee_to_df(
        feature_collection,
        external_id_column,
        national_codes=national_codes,
        unit_type=unit_type,
        whisp_image=whisp_image,
        custom_bands=custom_bands,  # Pass through
    )


def whisp_formatted_stats_geojson_to_df(
    input_geojson_filepath: Path | str,
    external_id_column=None,
    remove_geom=False,
    national_codes=None,
    unit_type="ha",
    whisp_image=None,
    custom_bands=None,
    mode: str = "sequential",
    batch_size: int = 10,
    max_concurrent: int = 20,
    geometry_audit_trail: bool = False,
) -> pd.DataFrame:
    """
    Main entry point for converting GeoJSON to Whisp statistics.

    Routes to the appropriate processing mode with automatic formatting and validation.

    Converts a GeoJSON file to a pandas DataFrame containing Whisp stats for the input ROI.
    Output DataFrame is validated against a Panderas schema (created from lookup CSVs).
    Results are automatically formatted and unit-converted (ha or percent).

    If `external_id_column` is provided, it will be used to link external identifiers
    from the input GeoJSON to the output DataFrame.

    Parameters
    ----------
    input_geojson_filepath : Path | str
        The filepath to the GeoJSON of the ROI to analyze.
    external_id_column : str, optional
        The column in the GeoJSON containing external IDs to be preserved in the output DataFrame.
        This column must exist as a property in ALL features of the GeoJSON file.
        Use debug_feature_collection_properties() to inspect available properties if you encounter errors.
    national_codes : list, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    whisp_image : ee.Image, optional
        Pre-combined multiband Earth Engine Image containing all Whisp datasets.
        If provided, this image will be used instead of combining datasets based on national_codes.
        If None, datasets will be combined automatically using national_codes parameter.
    custom_bands : list or dict, optional
        Custom band information for extra columns. Can be:
        - List of band names: ['Aa_test', 'elevation']
        - Dict with types: {'Aa_test': 'float64', 'elevation': 'float32'}
        - None: preserves all extra columns automatically
    mode : str, optional
        Processing mode, by default "concurrent":
        - "concurrent": Uses high-volume endpoint with concurrent batching (recommended for large files)
        - "sequential": Uses standard endpoint for sequential processing (more stable)
        - "legacy": Uses original implementation (basic stats extraction only, no formatting)
    batch_size : int, optional
        Features per batch for concurrent/sequential modes, by default 10.
        Only applicable for "concurrent" and "sequential" modes.
    max_concurrent : int, optional
        Maximum concurrent EE calls for concurrent mode, by default 20.
        Only applicable for "concurrent" mode.
    geometry_audit_trail : bool, default True
        If True (default), includes audit trail columns:
        - geo_original: Original input geometry
        - geometry_type_original: Original geometry type
        - geometry_type: Processed geometry type (from EE)
        - geometry_type_changed: Boolean flag if geometry changed
        - geometry_degradation_type: Description of how it changed

        Processing metadata stored in df.attrs['processing_metadata'].
        These columns enable full transparency for geometry modifications during processing.

    Returns
    -------
    df_stats : pd.DataFrame
        The DataFrame containing the Whisp stats for the input ROI,
        automatically formatted and validated.

    Examples
    --------
    >>> # Use concurrent processing (default, recommended for large datasets)
    >>> df = whisp_formatted_stats_geojson_to_df("data.geojson")

    >>> # Use sequential processing for more stable/predictable results
    >>> df = whisp_formatted_stats_geojson_to_df(
    ...     "data.geojson",
    ...     mode="sequential"
    ... )

    >>> # Adjust concurrency parameters
    >>> df = whisp_formatted_stats_geojson_to_df(
    ...     "large_data.geojson",
    ...     mode="concurrent",
    ...     max_concurrent=30,
    ...     batch_size=15
    ... )

    >>> # Use legacy mode for backward compatibility (basic extraction only)
    >>> df = whisp_formatted_stats_geojson_to_df(
    ...     "data.geojson",
    ...     mode="legacy"
    ... )
    """
    # Import here to avoid circular imports
    try:
        from openforis_whisp.advanced_stats import (
            whisp_formatted_stats_geojson_to_df_fast,
        )
    except ImportError:
        # Fallback to legacy if advanced_stats not available
        mode = "legacy"

    logger = logging.getLogger("whisp")

    if mode == "legacy":
        # Log info if batch_size or max_concurrent were passed but won't be used
        if batch_size != 10 or max_concurrent != 20:
            unused = []
            if batch_size != 10:
                unused.append(f"batch_size={batch_size}")
            if max_concurrent != 20:
                unused.append(f"max_concurrent={max_concurrent}")
            logger.info(
                f"Mode is 'legacy': {', '.join(unused)}\n"
                "parameter(s) are not used in legacy mode."
            )
        # Use original implementation (basic stats extraction only)
        return whisp_formatted_stats_geojson_to_df_legacy(
            input_geojson_filepath=input_geojson_filepath,
            external_id_column=external_id_column,
            national_codes=national_codes,
            unit_type=unit_type,
            whisp_image=whisp_image,
            custom_bands=custom_bands,
        )
    elif mode in ("concurrent", "sequential"):
        # Log info if batch_size or max_concurrent are not used in sequential mode
        if mode == "sequential":
            unused = []
            if batch_size != 10:
                unused.append(f"batch_size={batch_size}")
            if max_concurrent != 20:
                unused.append(f"max_concurrent={max_concurrent}")
            if unused:
                logger.info(
                    f"Mode is 'sequential': {', '.join(unused)}\n"
                    "parameter(s) are not used in sequential (single-threaded) mode."
                )
        # Route to fast function with explicit mode (skip auto-detection)
        return whisp_formatted_stats_geojson_to_df_fast(
            input_geojson_filepath=input_geojson_filepath,
            external_id_column=external_id_column,
            national_codes=national_codes,
            unit_type=unit_type,
            whisp_image=whisp_image,
            custom_bands=custom_bands,
            mode=mode,  # Pass mode directly (concurrent or sequential)
            batch_size=batch_size,
            max_concurrent=max_concurrent,
            geometry_audit_trail=geometry_audit_trail,
        )
    else:
        raise ValueError(
            f"Invalid mode '{mode}'. Must be 'concurrent', 'sequential', or 'legacy'."
        )


def whisp_formatted_stats_geojson_to_geojson(
    input_geojson_filepath,
    output_geojson_filepath,
    external_id_column=None,
    geo_column: str = "geo",
    national_codes=None,
    unit_type="ha",
    whisp_image=None,  # New parameter
):
    """
    Convert a formatted GeoJSON file with a geo column into a GeoJSON file containing Whisp stats.

    Parameters
    ----------
    input_geojson_filepath : str
        The filepath to the input GeoJSON file.
    output_geojson_filepath : str
        The filepath to save the output GeoJSON file.
    external_id_column : str, optional
        The name of the column containing external IDs, by default None.
    geo_column : str, optional
        The name of the column containing GeoJSON geometries, by default "geo".
    national_codes : list, optional
        List of ISO2 country codes to include national datasets.
    unit_type : str, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    whisp_image : ee.Image, optional
        Pre-combined multiband Earth Engine Image containing all Whisp datasets.

    Returns
    -------
    None
    """
    df = whisp_formatted_stats_geojson_to_df(
        input_geojson_filepath=input_geojson_filepath,
        external_id_column=external_id_column,
        national_codes=national_codes,
        unit_type=unit_type,
        whisp_image=whisp_image,  # Pass through
    )
    # Convert the df to GeoJSON
    convert_df_to_geojson(df, output_geojson_filepath, geo_column)

    # Suppress verbose output
    # print(f"GeoJSON with Whisp stats saved to {output_geojson_filepath}")


def whisp_formatted_stats_ee_to_geojson(
    feature_collection: ee.FeatureCollection,
    output_geojson_filepath: str,
    external_id_column=None,
    geo_column: str = "geo",
    national_codes=None,
    unit_type="ha",
    whisp_image=None,  # New parameter
):
    """
    Convert an Earth Engine FeatureCollection to a GeoJSON file containing Whisp stats.

    Parameters
    ----------
    feature_collection : ee.FeatureCollection
        The feature collection of the ROI to analyze.
    output_geojson_filepath : str
        The filepath to save the output GeoJSON file.
    external_id_column : str, optional
        The name of the column containing external IDs, by default None.
    geo_column : str, optional
        The name of the column containing GeoJSON geometries, by default "geo".
    national_codes : list, optional
        List of ISO2 country codes to include national datasets.
    unit_type : str, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    whisp_image : ee.Image, optional
        Pre-combined multiband Earth Engine Image containing all Whisp datasets.
    Returns
    -------
    None
    """
    # Convert ee feature collection to a pandas dataframe
    df_stats = whisp_formatted_stats_ee_to_df(
        feature_collection,
        external_id_column,
        national_codes=national_codes,
        unit_type=unit_type,
        whisp_image=whisp_image,  # Pass through
    )

    # Convert the df to GeoJSON
    convert_df_to_geojson(df_stats, output_geojson_filepath, geo_column)

    print(f"GeoJSON with Whisp stats saved to {output_geojson_filepath}")


def whisp_formatted_stats_ee_to_df(
    feature_collection: ee.FeatureCollection,
    external_id_column=None,
    remove_geom=False,
    national_codes=None,
    unit_type="ha",
    whisp_image=None,
    custom_bands=None,  # New parameter
) -> pd.DataFrame:
    """
    Convert a feature collection to a validated DataFrame with Whisp statistics.

    Parameters
    ----------
    feature_collection : ee.FeatureCollection
        The feature collection of the ROI to analyze.
    external_id_column : str, optional
        The name of the external ID column, by default None.
    remove_geom : bool, optional
        Whether to remove the geometry column, by default False.
    national_codes : list, optional
        List of ISO2 country codes to include national datasets.
    unit_type : str, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    whisp_image : ee.Image, optional
        Pre-combined multiband Earth Engine Image containing all Whisp datasets.
    custom_bands : list or dict, optional
        Custom band information for extra columns.

    Returns
    -------
    validated_df : pd.DataFrame
        The validated dataframe containing the Whisp stats for the input ROI.
    """
    # Convert ee feature collection to a pandas dataframe
    df_stats = whisp_stats_ee_to_df(
        feature_collection,
        external_id_column,
        remove_geom,
        national_codes=national_codes,
        unit_type=unit_type,
        whisp_image=whisp_image,
    )

    # Use flexible validation that handles custom bands
    validated_df = validate_dataframe_using_lookups_flexible(
        df_stats, national_codes=national_codes, custom_bands=custom_bands
    )
    return validated_df


### functions without additional formatting below (i.e., raw output from GEE processing without schema validation step)


def whisp_stats_geojson_to_df(
    input_geojson_filepath: Path | str,
    external_id_column=None,
    national_codes=None,
    unit_type="ha",
    whisp_image=None,  # New parameter
) -> pd.DataFrame:
    """
    Convert a GeoJSON file to a pandas DataFrame with Whisp statistics.

    Parameters
    ----------
    input_geojson_filepath : Path | str
        The filepath to the GeoJSON of the ROI to analyze.
    external_id_column : str, optional
        The name of the external ID column, by default None.
    remove_geom : bool, optional
        Whether to remove the geometry column, by default False.
    national_codes : list, optional
        List of ISO2 country codes to include national datasets.
    unit_type : str, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    whisp_image : ee.Image, optional
        Pre-combined multiband Earth Engine Image containing all Whisp datasets.

    Returns
    -------
    df_stats : pd.DataFrame
        The dataframe containing the Whisp stats for the input ROI.
    """
    feature_collection = convert_geojson_to_ee(str(input_geojson_filepath))

    return whisp_stats_ee_to_df(
        feature_collection,
        external_id_column,
        national_codes=national_codes,
        unit_type=unit_type,
        whisp_image=whisp_image,  # Pass through
    )


def whisp_stats_geojson_to_ee(
    input_geojson_filepath: Path | str,
    external_id_column=None,
    national_codes=None,
    whisp_image=None,  # New parameter
) -> ee.FeatureCollection:
    """
    Convert a GeoJSON file to an Earth Engine FeatureCollection with Whisp statistics.

    Parameters
    ----------
    input_geojson_filepath : Path | str
        The filepath to the GeoJSON of the ROI to analyze.
    external_id_column : str, optional
        The name of the external ID column, by default None.
    national_codes : list, optional
        List of ISO2 country codes to include national datasets.
    whisp_image : ee.Image, optional
        Pre-combined multiband Earth Engine Image containing all Whisp datasets.

    Returns
    -------
    ee.FeatureCollection
        The feature collection containing the Whisp stats for the input ROI.
    """
    feature_collection = convert_geojson_to_ee(str(input_geojson_filepath))

    return whisp_stats_ee_to_ee(
        feature_collection,
        external_id_column,
        national_codes=national_codes,
        whisp_image=whisp_image,  # Pass through
    )


def whisp_stats_geojson_to_geojson(
    input_geojson_filepath,
    output_geojson_filepath,
    external_id_column=None,
    national_codes=None,
    unit_type="ha",
    whisp_image=None,  # New parameter
):
    """
    Convert a GeoJSON file to a GeoJSON object containing Whisp stats for the input ROI.

    Parameters
    ----------
    input_geojson_filepath : str
        The filepath to the input GeoJSON file.
    output_geojson_filepath : str
        The filepath to save the output GeoJSON file.
    external_id_column : str, optional
        The name of the column containing external IDs, by default None.
    national_codes : list, optional
        List of ISO2 country codes to include national datasets.
    unit_type : str, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    whisp_image : ee.Image, optional
        Pre-combined multiband Earth Engine Image containing all Whisp datasets.

    Returns
    -------
    None
    """
    # Convert GeoJSON to Earth Engine FeatureCollection
    feature_collection = convert_geojson_to_ee(input_geojson_filepath)

    # Get stats as a FeatureCollection
    stats_feature_collection = whisp_stats_ee_to_ee(
        feature_collection,
        external_id_column,
        national_codes=national_codes,
        unit_type=unit_type,
        whisp_image=whisp_image,  # Pass through
    )

    # Convert the stats FeatureCollection to GeoJSON
    stats_geojson = convert_ee_to_geojson(stats_feature_collection)

    # Save the GeoJSON to a file
    with open(output_geojson_filepath, "w") as f:
        json.dump(stats_geojson, f, indent=2)


def whisp_stats_geojson_to_drive(
    input_geojson_filepath: Path | str,
    external_id_column=None,
    national_codes=None,
    unit_type="ha",
    whisp_image=None,  # New parameter
):
    """
    Export Whisp statistics for a GeoJSON file to Google Drive.

    Parameters
    ----------
    input_geojson_filepath : Path | str
        The filepath to the GeoJSON of the ROI to analyze.
    external_id_column : str, optional
        The name of the external ID column, by default None.
    national_codes : list, optional
        List of ISO2 country codes to include national datasets.
    unit_type : str, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    whisp_image : ee.Image, optional
        Pre-combined multiband Earth Engine Image containing all Whisp datasets.

    Returns
    -------
    Message showing location of file in Google Drive
    """
    try:
        input_geojson_filepath = Path(input_geojson_filepath)
        if not input_geojson_filepath.exists():
            raise FileNotFoundError(f"File {input_geojson_filepath} does not exist.")

        feature_collection = convert_geojson_to_ee(str(input_geojson_filepath))

        return whisp_stats_ee_to_drive(
            feature_collection,
            external_id_column,
            national_codes=national_codes,
            unit_type=unit_type,
            whisp_image=whisp_image,  # Pass through
        )

    except Exception as e:
        print(f"An error occurred: {e}")


def whisp_stats_ee_to_ee(
    feature_collection,
    external_id_column,
    national_codes=None,
    unit_type="ha",
    keep_properties=None,
    whisp_image=None,
    validate_external_id=True,
    validate_bands=False,  # New parameter
):
    """
    Process a feature collection to get statistics for each feature.

    Parameters:
        feature_collection (ee.FeatureCollection): The input feature collection.
        external_id_column (str): The name of the external ID column to check.
        national_codes (list, optional): List of ISO2 country codes to include national datasets.
        unit_type (str): Whether to use hectares ("ha") or percentage ("percent"), default "ha".
        keep_properties (None, bool, or list, optional): Properties to keep from the input features.
            - None: Remove all properties (default behavior)
            - True: Keep all properties
            - list: Keep only the specified properties
        whisp_image (ee.Image, optional): Pre-combined multiband Earth Engine Image containing
            all Whisp datasets. If provided, this image will be used instead of combining
            datasets based on national_codes.
        validate_external_id (bool, optional): If True, validates that external_id_column exists
            in all features (default: True). Set to False to skip validation and save 2-4 seconds.
            Only disable if you're confident the column exists in all features.

    Returns:
        ee.FeatureCollection: The output feature collection with statistics.
    """
    if external_id_column is not None:
        try:
            # OPTIMIZATION: Make validation optional to save 2-4 seconds
            # Validation includes multiple .getInfo() calls which are slow
            if validate_external_id:
                # Validate that the external_id_column exists in all features
                validation_result = validate_external_id_column(
                    feature_collection, external_id_column
                )

                if not validation_result["is_valid"]:
                    raise ValueError(validation_result["error_message"])

            # First handle property selection, but preserve the external_id_column
            if keep_properties is not None:
                if keep_properties == True:
                    # Keep all properties including external_id_column
                    pass  # No need to modify feature_collection
                elif isinstance(keep_properties, list):
                    # Ensure external_id_column is included in the list
                    if external_id_column not in keep_properties:
                        keep_properties = keep_properties + [external_id_column]
                    feature_collection = feature_collection.select(keep_properties)
                else:
                    raise ValueError(
                        "keep_properties must be None, True, or a list of property names."
                    )

            # Set the external_id with robust null handling
            def set_external_id_safely_and_clean(feature):
                external_id_value = feature.get(external_id_column)
                # Use server-side null checking and string conversion
                external_id_value = ee.Algorithms.If(
                    ee.Algorithms.IsEqual(external_id_value, None),
                    "unknown",
                    ee.String(external_id_value),
                )
                # Create a new feature with the standardized external_id column
                # Note: we use "external_id" as the standardized column name, not the original external_id_column name
                return ee.Feature(feature.set("external_id", external_id_value))

            feature_collection = feature_collection.map(
                set_external_id_safely_and_clean
            )

            # Finally, clean up to keep only geometry and external_id if keep_properties is None
            if keep_properties is None:
                feature_collection = feature_collection.select(["external_id"])

        except Exception as e:
            # Handle the exception and provide a helpful error message
            print(
                f"An error occurred when trying to set the external_id_column: {external_id_column}. Error: {e}"
            )
            raise e  # Re-raise the exception to stop execution
    else:
        feature_collection = _keep_fc_properties(feature_collection, keep_properties)

    fc = get_stats(
        feature_collection,
        national_codes=national_codes,
        unit_type=unit_type,
        whisp_image=whisp_image,  # Pass through
        validate_bands=validate_bands,
    )

    return add_id_to_feature_collection(dataset=fc, id_name=plot_id_column)


def _keep_fc_properties(feature_collection, keep_properties):
    """
    Filter feature collection properties based on keep_properties parameter.

    OPTIMIZATION: When keep_properties is True, we no longer call .getInfo()
    to get property names. Instead, we simply return the collection as-is,
    since True means "keep all properties". This saves 1-2 seconds.
    """
    # If keep_properties is specified, select only those properties
    if keep_properties is None:
        feature_collection = feature_collection.select([])
    elif keep_properties == True:
        # If keep_properties is true, keep all properties
        # No need to call .select() or .getInfo() - just return as-is
        pass
    elif isinstance(keep_properties, list):
        feature_collection = feature_collection.select(keep_properties)
    else:
        raise ValueError(
            "keep_properties must be None, True, or a list of property names."
        )
    return feature_collection


def whisp_stats_ee_to_df(
    feature_collection: ee.FeatureCollection,
    external_id_column=None,
    remove_geom=False,
    national_codes=None,
    unit_type="ha",
    whisp_image=None,
    validate_bands=False,  # New parameter
) -> pd.DataFrame:
    """
    Convert a Google Earth Engine FeatureCollection to a pandas DataFrame and convert ISO3 to ISO2 country codes.

    Parameters
    ----------
    feature_collection : ee.FeatureCollection
        The input FeatureCollection to analyze.
    external_id_column : str, optional
        The name of the external ID column, by default None.
    remove_geom : bool, optional
        Whether to remove the geometry column, by default True.
    national_codes : list, optional
        List of ISO2 country codes to include national datasets.
    unit_type : str, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    whisp_image : ee.Image, optional
        Pre-combined multiband Earth Engine Image containing all Whisp datasets.

    Returns
    -------
    df_stats : pd.DataFrame
        The dataframe containing the Whisp stats for the input ROI.
    """
    # First, do the whisp processing to get the EE feature collection with stats
    try:
        try:
            stats_feature_collection = whisp_stats_ee_to_ee(
                feature_collection,
                external_id_column,
                national_codes=national_codes,
                unit_type=unit_type,
                whisp_image=whisp_image,  # Pass through
                validate_bands=False,  # try withoutb validation first
            )
        except Exception as e:
            print(f"An error occurred during Whisp stats processing: {e}")
            raise e

        # Then, convert the EE feature collection to DataFrame
        try:
            df_stats = convert_ee_to_df(
                ee_object=stats_feature_collection,
                remove_geom=remove_geom,
            )
        except Exception as e:
            print(f"An error occurred during the conversion from EE to DataFrame: {e}")
            raise e

    except:  # retry with validation of whisp input datasets
        try:
            stats_feature_collection = whisp_stats_ee_to_ee(
                feature_collection,
                external_id_column,
                national_codes=national_codes,
                unit_type=unit_type,
                whisp_image=whisp_image,
                validate_bands=True,  # If error, try with validation
            )
        except Exception as e:
            print(f"An error occurred during Whisp stats processing: {e}")
            raise e

        # Then, convert the EE feature collection to DataFrame
        try:
            df_stats = convert_ee_to_df(
                ee_object=stats_feature_collection,
                remove_geom=remove_geom,
            )
        except Exception as e:
            print(f"An error occurred during the conversion from EE to DataFrame: {e}")
            raise e
    try:
        df_stats = convert_iso3_to_iso2(
            df=df_stats,
            iso3_column=iso3_country_column,
            iso2_column=iso2_country_column,
        )
    except Exception as e:
        print(f"An error occurred during the ISO3 to ISO2 conversion: {e}")
        return pd.DataFrame()  # Return an empty DataFrame in case of error

    # NEW: Set area to 0 for point geometries
    try:
        df_stats = set_point_geometry_area_to_zero(df_stats)
    except Exception as e:
        print(f"An error occurred during point geometry area adjustment: {e}")
        # Continue without the adjustment rather than failing completely

    # Reformat geometry types (MultiPolygon -> Polygon)
    try:
        df_stats = reformat_geometry_type(df_stats)
    except Exception as e:
        print(f"An error occurred during geometry type reformatting: {e}")
        # Continue without the adjustment rather than failing completely

    return df_stats


def set_point_geometry_area_to_zero(df: pd.DataFrame) -> pd.DataFrame:
    """
    Set the geometry area column to 0 for features with Point geometry type.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing geometry type and area columns

    Returns
    -------
    pd.DataFrame
        DataFrame with area set to 0 for Point geometries
    """
    # Check if required columns exist
    if geometry_type_column not in df.columns:
        print(
            f"Warning: {geometry_type_column} column not found. Skipping area adjustment for points."
        )
        return df

    # Create a copy to avoid modifying the original
    df_modified = df.copy()

    # Set area to 0 where geometry type is Point
    point_mask = df_modified[geometry_type_column] == "Point"
    df_modified.loc[point_mask, geometry_area_column] = 0.0

    # Log the changes
    num_points = point_mask.sum()
    if num_points > 0:
        print(f"Set area to 0 for {num_points} Point geometries")

    return df_modified


def reformat_geometry_type(df: pd.DataFrame) -> pd.DataFrame:
    """
    Reformat geometry type classification in the DataFrame output.
    Standardizes MultiPolygon geometry type to Polygon for consistent output.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing geometry type column

    Returns
    -------
    pd.DataFrame
        DataFrame with standardized geometry types
    """
    # Check if required columns exist
    if geometry_type_column not in df.columns:
        print(
            f"Warning: {geometry_type_column} column not found. Skipping geometry type reformatting."
        )
        return df

    # Create a copy to avoid modifying the original
    df_modified = df.copy()

    # Reformat MultiPolygon to Polygon
    multipolygon_mask = df_modified[geometry_type_column] == "MultiPolygon"
    df_modified.loc[multipolygon_mask, geometry_type_column] = "Polygon"

    # Log the changes
    num_reformatted = multipolygon_mask.sum()
    # if num_reformatted > 0:
    #     print(f"Reformatted {num_reformatted} MultiPolygon geometries to Polygon")

    return df_modified


def whisp_stats_ee_to_drive(
    feature_collection: ee.FeatureCollection,
    external_id_column=None,
    national_codes=None,
    unit_type="ha",
    whisp_image=None,  # New parameter
):
    """
     Export Whisp statistics for a feature collection to Google Drive.

     Parameters
     ----------
     feature_collection : ee.FeatureCollection
         The feature collection to analyze.
     external_id_column : str, optional
         The name of the external ID column, by default None.
     national_codes : list, optional
         List of ISO2 country codes to include national datasets.
    unit_type : str, optional
         Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    whisp_image : ee.Image, optional
         Pre-combined multiband Earth Engine Image containing all Whisp datasets.
     Returns
     -------
     None
    """
    try:
        task = ee.batch.Export.table.toDrive(
            collection=whisp_stats_ee_to_ee(
                feature_collection,
                external_id_column,
                national_codes=national_codes,
                unit_type=unit_type,
                whisp_image=whisp_image,  # Pass through
            ),
            description="whisp_output_table",
            # folder="whisp_results",
            fileFormat="CSV",
        )
        task.start()
        print(
            "Exporting to Google Drive: 'whisp_output_table.csv'. To track progress: https://code.earthengine.google.com/tasks"
        )
    except Exception as e:
        print(f"An error occurred during the export: {e}")


#### main stats functions


# Get stats for a feature or feature collection
def get_stats(
    feature_or_feature_col,
    national_codes=None,
    unit_type="ha",
    whisp_image=None,
    validate_bands=False,
):
    """
    Get stats for a feature or feature collection with optional pre-combined image.

    Parameters
    ----------
    feature_or_feature_col : ee.Feature or ee.FeatureCollection
        The input feature or feature collection to analyze
    national_codes : list, optional
        List of ISO2 country codes to include national datasets.
        Only used if whisp_image is None.
    unit_type : str, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    whisp_image : ee.Image, optional
        Pre-combined multiband Earth Engine Image containing all Whisp datasets.
        If provided, this will be used instead of combining datasets based on national_codes.
        If None, datasets will be combined automatically using national_codes parameter.
    Returns
    -------
    ee.FeatureCollection
        Feature collection with calculated statistics
    """

    # Use provided image or combine datasets
    if whisp_image is not None:
        img_combined = whisp_image
        print("Using provided whisp_image")
    else:
        img_combined = combine_datasets(
            national_codes=national_codes,
            validate_bands=validate_bands,
            include_context_bands=False,
        )
        print(f"Combining datasets with national_codes: {national_codes}")

    # Check if the input is a Feature or a FeatureCollection
    if isinstance(feature_or_feature_col, ee.Feature):
        print("Processing single feature")
        # OPTIMIZATION: Create cached images for single feature processing
        water_all = get_water_flag_image()
        bounds_ADM1 = get_admin_boundaries_fc()
        output = ee.FeatureCollection(
            [
                get_stats_feature(
                    feature_or_feature_col,
                    img_combined,
                    unit_type=unit_type,
                    water_all=water_all,
                    bounds_ADM1=bounds_ADM1,
                )
            ]
        )
    elif isinstance(feature_or_feature_col, ee.FeatureCollection):
        print("Processing feature collection")
        output = get_stats_fc(
            feature_or_feature_col,
            national_codes=national_codes,
            unit_type=unit_type,
            img_combined=img_combined,  # Pass the image directly
        )
    else:
        output = "Check inputs: not an ee.Feature or ee.FeatureCollection"
    return output


# Get statistics for a feature collection
def get_stats_fc(feature_col, national_codes=None, unit_type="ha", img_combined=None):
    """
    Calculate statistics for a feature collection using Whisp datasets.

    OPTIMIZATION: Creates water flag and admin_boundaries images once and reuses
    them for all features instead of recreating them for each feature.
    This saves 7-15 seconds per analysis.

    Parameters
    ----------
    feature_col : ee.FeatureCollection
        The input feature collection to analyze
    national_codes : list, optional
        List of ISO2 country codes (e.g., ["BR", "US"]) to include national datasets.
        If provided, only national datasets for these countries and global datasets will be used.
        If None (default), only global datasets will be used.
        Only used if img_combined is None.
    unit_type : str, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    img_combined : ee.Image, optional
        Pre-combined multiband image containing all Whisp datasets.
        If provided, this will be used instead of combining datasets based on national_codes.
    Returns
    -------
    ee.FeatureCollection
        Feature collection with calculated statistics
    """
    # OPTIMIZATION: Create cached images once before processing features
    # These will be reused for all features instead of being recreated each time
    water_all = get_water_flag_image()
    bounds_ADM1 = get_admin_boundaries_fc()

    out_feature_col = ee.FeatureCollection(
        feature_col.map(
            lambda feature: get_stats_feature(
                feature,
                img_combined,
                unit_type=unit_type,
                water_all=water_all,
                bounds_ADM1=bounds_ADM1,
            )
        )
    )
    # print(out_feature_col.first().getInfo()) # for testing

    return out_feature_col


# Get statistics for a single feature
# Note: This function doesn't need whisp_image parameter since it already accepts img_combined directly


def get_stats_feature(
    feature, img_combined, unit_type="ha", water_all=None, bounds_ADM1=None
):
    """
    Get statistics for a single feature using a pre-combined image.

    OPTIMIZATION: Accepts cached water/admin_boundaries images to avoid recreating
    them for every feature.

    Parameters
    ----------
    feature : ee.Feature
        The feature to analyze
    img_combined : ee.Image
        Pre-combined image with all the datasets
    unit_type : str, optional
        Whether to use hectares ("ha") or percentage ("percent"), by default "ha".
    water_all : ee.Image, optional
        Cached water flag image
    bounds_ADM1 : ee.FeatureCollection, optional
        Cached admin_boundaries feature collection

    Returns
    -------
    ee.Feature
        Feature with calculated statistics
    """
    reduce = img_combined.reduceRegion(
        reducer=ee.Reducer.sum(),
        geometry=feature.geometry(),
        scale=10,
        maxPixels=1e10,
        tileScale=8,
    )

    # Get basic feature information with cached images
    feature_info = get_type_and_location(feature, water_all, bounds_ADM1)

    # add statistics unit type (e.g., percentage or hectares) to dictionary
    stats_unit_type = ee.Dictionary({stats_unit_type_column: unit_type})

    # Now, modified_dict contains all keys with the prefix added
    reduce_ha = reduce.map(
        lambda key, val: divide_and_format(ee.Number(val), ee.Number(10000))
    )

    # Get value for hectares
    area_ha = ee.Number(ee.Dictionary(reduce_ha).get(geometry_area_column))

    # Apply the function to each value in the dictionary using map()
    reduce_percent = reduce_ha.map(
        lambda key, val: percent_and_format(ee.Number(val), area_ha)
    )

    # Reformat the hectare statistics
    reducer_stats_ha = reduce_ha.set(
        geometry_area_column, area_ha.format(geometry_area_column_formatting)
    )  # area ha (formatted)

    # Reformat the percentage statistics
    reducer_stats_percent = reduce_percent.set(
        geometry_area_column, area_ha.format(geometry_area_column_formatting)
    )  # area ha (formatted)

    # Add country info onto hectare analysis results
    properties_ha = feature_info.combine(ee.Dictionary(reducer_stats_ha)).combine(
        stats_unit_type
    )

    # Add country info onto percentage analysis results
    properties_percent = feature_info.combine(
        ee.Dictionary(reducer_stats_percent)
    ).combine(stats_unit_type)

    # Choose whether to use hectares or percentage based on the parameter instead of global variable
    out_feature = ee.Algorithms.If(
        unit_type == "ha",
        feature.set(properties_ha),  # .setGeometry(None),
        feature.set(properties_percent),  # .setGeometry(None),
    )

    return out_feature


# Get basic feature information - uses admin and water datasets in gee.
def get_type_and_location(feature, water_all=None, bounds_ADM1=None):
    """
    Extracts basic feature information including country, admin area, geometry type, coordinates, and water flags.

    OPTIMIZATION: Accepts cached water flag image and admin_boundaries collection
    to avoid recreating them for every feature (saves 7-15 seconds per analysis).

    Parameters
    ----------
    feature : ee.Feature
        The feature to extract information from
    water_all : ee.Image, optional
        Cached water flag image. If None, creates it.
    bounds_ADM1 : ee.FeatureCollection, optional
        Cached admin_boundaries feature collection. If None, loads it.

    Returns
    -------
    ee.Dictionary
        Dictionary with feature information
    """
    # Get centroid of the feature's geometry
    centroid = feature.geometry().centroid(0.1)

    # OPTIMIZATION: Use cached admin_boundaries
    if bounds_ADM1 is None:
        bounds_ADM1 = get_admin_boundaries_fc()

    # Fetch location info from GAUL 2024 L1 (country, admin)
    location = ee.Dictionary(get_admin_boundaries_info(centroid, bounds_ADM1))
    country = ee.Dictionary({iso3_country_column: location.get("iso3_code")})

    admin_1 = ee.Dictionary(
        {admin_1_column: location.get("gaul1_name")}
    )  # Administrative level 1 (from GAUL 2024 L1)

    # OPTIMIZATION: Use cached water flag image
    if water_all is None:
        water_all = get_water_flag_image()

    # OPTIMIZATION: Use cached water flag image
    if water_all is None:
        water_all = get_water_flag_image()

    # Prepare the water flag information
    water_flag_dict = value_at_point_flag(
        point=centroid, image=water_all, band_name=water_flag, output_name=water_flag
    )

    # Get the geometry type of the feature
    geom_type = ee.Dictionary({geometry_type_column: feature.geometry().type()})

    # Get the coordinates (latitude, longitude) of the centroid
    coords_list = centroid.coordinates()
    coords_dict = ee.Dictionary(
        {
            centroid_x_coord_column: ee.Number(coords_list.get(0)).format(
                "%.6f"
            ),  # Longitude (6 dp)
            centroid_y_coord_column: ee.Number(coords_list.get(1)).format(
                "%.6f"
            ),  # Latitude (6 dp)
        }
    )

    # Combine all the extracted info into a single dictionary
    feature_info = (
        country.combine(admin_1)
        .combine(geom_type)
        .combine(coords_dict)
        .combine(water_flag_dict)
    )

    return feature_info


# Define a function to divide each value by 10,000 and format it with one decimal place
def divide_and_format(val, unit):
    # Convert the image to an ee.Number, divide by 10,000, and format with one decimal place
    formatted_value = ee.Number.parse(
        ee.Number(ee.Number(val).divide(ee.Number(unit))).format(
            stats_area_columns_formatting
        )
    )
    # Return the formatted value
    return ee.Number(formatted_value)


# Define a function to divide by total area of geometry and multiply by 100
def percent_and_format(val, area_ha):
    formatted_value = ee.Number.parse(
        ee.Number(ee.Number(val).divide(area_ha).multiply(ee.Number(100))).format(
            stats_percent_columns_formatting
        )
    )
    # Return the formatted value
    return ee.Number(formatted_value)


# GAUL 2024 L1 - admin units from FAO, allows commercial use
def get_admin_boundaries_info(geometry, bounds_ADM1=None):
    """
    Get GAUL 2024 L1 info for a geometry (country ISO3 code and admin boundary name).

    OPTIMIZATION: Accepts cached GAUL 2024 L1 FeatureCollection to avoid
    reloading it for every feature (saves 2-5 seconds per analysis).

    Parameters
    ----------
    geometry : ee.Geometry
        The geometry to query
    bounds_ADM1 : ee.FeatureCollection, optional
        Cached GAUL 2024 L1 feature collection. If None, loads it.

    Returns
    -------
    ee.Dictionary
        Dictionary with iso3_code (country) and gaul1_name (admin boundary name)
    """
    if bounds_ADM1 is None:
        bounds_ADM1 = get_admin_boundaries_fc()

    polygonsIntersectPoint = bounds_ADM1.filterBounds(geometry)
    backup_dict = ee.Dictionary({"iso3_code": "Unknown", "gaul1_name": "Unknown"})
    return ee.Algorithms.If(
        polygonsIntersectPoint.size().gt(0),
        polygonsIntersectPoint.first()
        .toDictionary()
        .select(["iso3_code", "gaul1_name"]),
        backup_dict,
    )


#####
# water flag - to flag plots that may be erroneous (i.e., where errors may have occured in their creation / translation  and so fall in either the ocean or inland water -
def usgs_gsv_ocean_prep():  # TO DO: for speed export image as an asset at samne res as JRC
    # Initialize the Earth Engine API
    # ee.Initialize()

    # Load the datasets
    mainlands = ee.FeatureCollection(
        "projects/sat-io/open-datasets/shoreline/mainlands"
    )
    big_islands = ee.FeatureCollection(
        "projects/sat-io/open-datasets/shoreline/big_islands"
    )
    small_islands = ee.FeatureCollection(
        "projects/sat-io/open-datasets/shoreline/small_islands"
    )

    # Combine the datasets into one FeatureCollection
    gsv = ee.FeatureCollection([mainlands, big_islands, small_islands]).flatten()

    # Rasterize the combined FeatureCollection and make areas outside coast (i.e. ocean) as value 1
    # and then rename the band
    return ee.Image(1).paint(gsv).selfMask().rename("ocean")


def jrc_water_surface_prep():
    jrc_surface_water = ee.Image("JRC/GSW1_4/GlobalSurfaceWater")

    # use transition band
    jrc_transition = jrc_surface_water.select("transition")

    # select permanent water bodies:
    # remap the following classes to have a value of 1:
    # "Permanent", "New Permanent", and "Seasonal to Permanent" (i.e., classes 1,2 and 7).
    # All other classes as value 0.
    permanent_inland_water = jrc_transition.remap([1, 2, 7], [1, 1, 1], 0).unmask()

    # optional - clip to within coast line (not needed currently and extra processing)
    # permanent_inland_water = permanent_inland_water.where(usgs_gsv_ocean_prep(),0)

    return permanent_inland_water.rename("water_inland")


def water_flag_all_prep():
    # combine both where water surface is 1, then 1, else use non_land_gsv
    return (
        usgs_gsv_ocean_prep()
        .unmask()
        .where(jrc_water_surface_prep(), 1)
        .rename(water_flag)
    )


def value_at_point_flag(point, image, band_name, output_name):
    """Sample an image at the given point and make a dictionary output where the name is defined by output_name parameter"""
    sample = image.sample(region=point, scale=30, numPixels=1).first()

    # Get the value from the sampled point
    value = sample.get(band_name)  # assuming the band name is 'b1', change if necessary

    # Use a conditional statement to check if the value is 1
    result = value  # ee.Algorithms.If(ee.Number(value).eq(1), "True", "False")

    # Return the output dictionary
    return ee.Dictionary({output_name: result})  # .getInfo()


def add_id_to_feature_collection(dataset, id_name):
    """
    Adds an incremental (1,2,3 etc) 'id' property to each feature in the given FeatureCollection.

    Args:
    - dataset: ee.FeatureCollection, the FeatureCollection to operate on.

    Returns:
    - dataset_with_id: ee.FeatureCollection, the FeatureCollection with 'id' property added to each feature.
    """
    # Get the list of system:index values
    indexes = dataset.aggregate_array("system:index")

    # Create a sequence of numbers starting from 1 to the size of indexes
    ids = ee.List.sequence(1, indexes.size())

    # Create a dictionary mapping system:index to id
    id_by_index = ee.Dictionary.fromLists(indexes, ids)

    # Function to add 'id' property to each feature
    def add_id(feature):
        # Get the system:index of the feature
        system_index = feature.get("system:index")

        # Get the id corresponding to the system:index
        feature_id = id_by_index.get(system_index)

        # Set the 'id' property of the feature
        return feature.set(id_name, feature_id)

    # Map the add_id function over the dataset
    dataset_with_id = dataset.map(add_id)

    return dataset_with_id


# Function to add ID to features
def add_id_to_feature(feature, id_name):
    index = feature.get("system:index")
    return feature.set(id_name, index)


# Function to flag positive values
def flag_positive_values(feature, flag_positive):
    for prop_name in flag_positive:
        flag_value = ee.Algorithms.If(
            ee.Number(feature.get(prop_name)).gt(0), "True", "-"
        )
        feature = feature.set(prop_name, flag_value)
    return feature


# Function to exclude properties
def copy_properties_and_exclude(feature, exclude_properties_from_output):
    return ee.Feature(feature.geometry()).copyProperties(
        source=feature, exclude=exclude_properties_from_output
    )


def ee_image_checker(image):
    """
    Tests if the input is a valid ee.Image.

    Args:
        image: An ee.Image object.

    Returns:
        bool: True if the input is a valid ee.Image, False otherwise.
    """
    try:
        if ee.Algorithms.ObjectType(image).getInfo() == "Image":
            # Trigger some action on the image to ensure it's a valid image
            image.getInfo()  # This will raise an exception if the image is invalid
            return True
    except ee.EEException as e:
        print(f"Image validation failed with EEException: {e}")
    except Exception as e:
        print(f"Image validation failed with exception: {e}")
    return False


def keep_valid_images(image_list):
    """
    Filters a list to return only valid ee.Images.

    Args:
        image_list: List of ee.Image objects.

    Returns:
        list: List of valid ee.Image objects.
    """
    valid_imgs = []
    for image in image_list:
        if ee_image_checker(image):
            valid_imgs.append(image)
    return valid_imgs


def convert_iso3_to_iso2(df, iso3_column, iso2_column):
    """
    Converts ISO3 country codes to ISO2 codes and adds a new column to the DataFrame.

    Args:
        df (pd.DataFrame): Input DataFrame containing ISO3 country codes.
        iso3_column (str): The column name in the DataFrame with ISO3 country codes.
        iso2_column (str): The new column name to store ISO2 country codes.

    Returns:
        pd.DataFrame: Updated DataFrame with the new ISO2 column.
    """
    import country_converter as coco

    # Apply conversion from ISO3 to ISO2
    df[iso2_column] = df[iso3_column].apply(
        lambda x: (
            coco.convert(names=x, to="ISO2") if x else "not found (disputed territory)"
        )
    )

    return df


def validate_external_id_column(feature_collection, external_id_column):
    """
    Validates that the external_id_column exists in all features of the collection.

    Parameters
    ----------
    feature_collection : ee.FeatureCollection
        The feature collection to validate
    external_id_column : str
        The name of the external ID column to check

    Returns
    -------
    dict
        Dictionary with validation results including:
        - 'is_valid': bool indicating if column exists in all features
        - 'total_features': int total number of features
        - 'features_with_column': int number of features that have the column
        - 'available_properties': list of properties available in first feature
        - 'error_message': str error message if validation fails
    """
    try:
        # Get total number of features
        total_features = feature_collection.size().getInfo()

        if total_features == 0:
            return {
                "is_valid": False,
                "total_features": 0,
                "features_with_column": 0,
                "available_properties": [],
                "error_message": "Feature collection is empty",
            }

        # Get available properties from first feature
        first_feature_props = feature_collection.first().propertyNames().getInfo()

        # Check if external_id_column exists in all features
        def check_column_exists(feature):
            has_column = feature.propertyNames().contains(external_id_column)
            return feature.set("_has_external_id", has_column)

        features_with_check = feature_collection.map(check_column_exists)
        features_with_column = (
            features_with_check.filter(ee.Filter.eq("_has_external_id", True))
            .size()
            .getInfo()
        )

        is_valid = features_with_column == total_features

        error_message = None
        if not is_valid:
            missing_count = total_features - features_with_column
            error_message = (
                f"The column '{external_id_column}' is missing from {missing_count} "
                f"out of {total_features} features in the collection. "
                f"Available properties in first feature: {first_feature_props}"
            )

        return {
            "is_valid": is_valid,
            "total_features": total_features,
            "features_with_column": features_with_column,
            "available_properties": first_feature_props,
            "error_message": error_message,
        }

    except Exception as e:
        return {
            "is_valid": False,
            "total_features": 0,
            "features_with_column": 0,
            "available_properties": [],
            "error_message": f"Error during validation: {str(e)}",
        }


def debug_feature_collection_properties(feature_collection, max_features=5):
    """
    Debug helper function to inspect the properties of features in a collection.

    Parameters
    ----------
    feature_collection : ee.FeatureCollection
        The feature collection to inspect
    max_features : int, optional
        Maximum number of features to inspect, by default 5

    Returns
    -------
    dict
        Dictionary with debugging information about the feature collection
    """
    try:
        total_features = feature_collection.size().getInfo()

        if total_features == 0:
            return {"total_features": 0, "error": "Feature collection is empty"}

        # Limit the number of features to inspect
        features_to_check = min(max_features, total_features)
        limited_fc = feature_collection.limit(features_to_check)

        # Get properties for each feature
        def get_feature_properties(feature):
            return ee.Dictionary(
                {
                    "properties": feature.propertyNames(),
                    "geometry_type": feature.geometry().type(),
                }
            )

        feature_info = limited_fc.map(get_feature_properties).getInfo()

        return {
            "total_features": total_features,
            "inspected_features": features_to_check,
            "feature_details": [
                {
                    "feature_index": i,
                    "properties": feature_info["features"][i]["properties"][
                        "properties"
                    ],
                    "geometry_type": feature_info["features"][i]["properties"][
                        "geometry_type"
                    ],
                }
                for i in range(len(feature_info["features"]))
            ],
        }

    except Exception as e:
        return {"error": f"Error during debugging: {str(e)}"}


# helper function to set area to 0 for point geometries
def set_point_geometry_area_to_zero(df: pd.DataFrame) -> pd.DataFrame:
    """
    Set the geometry area column to 0 for features with Point geometry type.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame containing geometry type and area columns

    Returns
    -------
    pd.DataFrame
        DataFrame with area set to 0 for Point geometries
    """
    # Check if required columns exist
    if geometry_type_column not in df.columns:
        print(
            f"Warning: {geometry_type_column} column not found. Skipping area adjustment for points."
        )
        return df

    if geometry_area_column not in df.columns:
        print(
            f"Warning: {geometry_area_column} column not found. Skipping area adjustment for points."
        )
        return df

    # Create a copy to avoid modifying the original
    df_modified = df.copy()

    # Set area to 0 where geometry type is Point
    point_mask = df_modified[geometry_type_column] == "Point"
    df_modified.loc[point_mask, geometry_area_column] = 0.0

    # Log the changes
    num_points = point_mask.sum()
    # if num_points > 0:
    #     print(f"Set area to 0 for {num_points} Point geometries")

    return df_modified



# This file contains python code for the Google Earth Engine datasets used in the Whisp pacakge.

# If you are running a bespoke analysis including your own datasets see also the main README.md file.

# Key aspects to include in the code for each function are:
# a) a suffix of ' _prep' and
# b) a prefix of "nXX_" if it is national/sub-national dataset (where XX is replaced by that country code), or a prefix of 'g_' if it covers more than one country.
# c) a name for your image, defined by ".rename('add_your_image_name_here')". This becomes the column header in the output table.

# Tips:
#  -Avoid getInfo() and for loops to speed up processing by keeping everything in the Earth Engine API.
#  -For all the above you will need to be running the package in editable mode for these local changes to take effect.
#   Editable mode runs the package locally and thus changes to any files are reflected immediately.

import ee

# ee.Authenticate()
# ee.Initialize()

from datetime import datetime

# from openforis_whisp.parameters.config_runtime import (
#     geometry_area_column,
# )  # ideally make relative import statement

# defining here instead of importing from config_runtime, to allow functioning as more of a standalone script
geometry_area_column = "Area"

# Calculate current year once at module load time (not in functions)
# This avoids repeated datetime calls and potential .getInfo() calls
CURRENT_YEAR = datetime.now().year
CURRENT_YEAR_2DIGIT = CURRENT_YEAR % 100  # Last two digits for RADD datasets

import inspect

import logging

# Configure logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


def get_logger(name):
    return logging.getLogger(name)


# Add datasets below

# tree cover datasets


# ESA_TC_2020
def g_esa_worldcover_trees_prep():
    esa_worldcover_2020_raw = ee.Image("ESA/WorldCover/v100/2020")
    esa_worldcover_trees_2020 = esa_worldcover_2020_raw.eq(95).Or(
        esa_worldcover_2020_raw.eq(10)
    )  # get trees and mnangroves
    return esa_worldcover_trees_2020.rename("ESA_TC_2020").selfMask()


# EUFO_2020
def g_jrc_gfc_2020_prep():
    # JRC GFC2020 V3 is a single Image with band 'Map'
    jrc_gfc2020 = ee.Image("JRC/GFC2020/V3").select("Map")
    return jrc_gfc2020.rename("EUFO_2020").selfMask()


# GFC_TC_2020
def g_glad_gfc_10pc_prep():
    gfc = ee.Image("UMD/hansen/global_forest_change_2024_v1_12")
    gfc_treecover2000 = gfc.select(["treecover2000"])
    gfc_loss2001_2020 = gfc.select(["lossyear"]).lte(20)
    gfc_treecover2020 = gfc_treecover2000.where(gfc_loss2001_2020.eq(1), 0)
    return gfc_treecover2020.gt(10).rename("GFC_TC_2020").selfMask()


# GLAD_Primary
def g_glad_pht_prep():
    primary_ht_forests2001_raw = ee.ImageCollection(
        "UMD/GLAD/PRIMARY_HUMID_TROPICAL_FORESTS/v1"
    )
    primary_ht_forests2001 = (
        primary_ht_forests2001_raw.select("Primary_HT_forests").mosaic().selfMask()
    )
    gfc = ee.Image("UMD/hansen/global_forest_change_2024_v1_12")
    gfc_loss2001_2020 = gfc.select(["lossyear"]).lte(20)
    return (
        primary_ht_forests2001.where(gfc_loss2001_2020.eq(1), 0)
        .rename("GLAD_Primary")
        .selfMask()
    )


# TMF_undist (undistrubed forest in 2020)
def g_jrc_tmf_undisturbed_prep():
    TMF_undist_2020 = (
        ee.ImageCollection("projects/JRC/TMF/v1_2024/AnnualChanges")
        .select("Dec2020")
        .mosaic()
        .eq(1)
    )  # update from https://github.com/forestdatapartnership/whisp/issues/42
    return TMF_undist_2020.rename("TMF_undist").selfMask()


# Forest Persistence FDaP
def g_fdap_forest_prep():
    fdap_forest_raw = ee.Image(
        "projects/forestdatapartnership/assets/community_forests/ForestPersistence_2020"
    )
    fdap_forest = fdap_forest_raw.gt(0.75)
    return fdap_forest.rename("Forest_FDaP").selfMask()


#########################primary forest
# EUFO JRC Global forest type - primary
def g_gft_primary_prep():
    gft_raw = ee.ImageCollection("JRC/GFC2020_subtypes/V0").mosaic()
    gft_primary = gft_raw.eq(10)
    return gft_primary.rename("GFT_primary").selfMask()


# Intact Forest Landscape 2020
def g_ifl_2020_prep():
    IFL_2020 = ee.Image("users/potapovpeter/IFL_2020")
    return IFL_2020.rename("IFL_2020").selfMask()


# European Primary Forest Dataset
def g_epfd_prep():
    EPFD = ee.FeatureCollection("HU_BERLIN/EPFD/V2/polygons")
    EPFD_binary = ee.Image().paint(EPFD, 1)
    return EPFD_binary.rename("European_Primary_Forest").selfMask()


# EUFO JRC Global forest type - naturally regenerating planted/plantation forests
def g_gft_nat_reg_prep():
    gft_raw = ee.ImageCollection("JRC/GFC2020_subtypes/V0").mosaic()
    gft_nat_reg = gft_raw.eq(1)
    return gft_nat_reg.rename("GFT_naturally_regenerating").selfMask()


#########################planted and plantation forests

# EUFO JRC Global forest type - planted/plantation forests
def g_gft_plantation_prep():
    gft_raw = ee.ImageCollection("JRC/GFC2020_subtypes/V0").mosaic()
    gft_plantation = gft_raw.eq(20)
    return gft_plantation.rename("GFT_planted_plantation").selfMask()


def g_iiasa_planted_prep():
    iiasa = ee.Image("projects/sat-io/open-datasets/GFM/FML_v3-2")
    iiasa_PL = iiasa.eq(31).Or(iiasa.eq(32))
    return iiasa_PL.rename("IIASA_planted_plantation").selfMask()


#########################TMF regrowth in 2023
def g_tmf_regrowth_prep():
    # Load the TMF Degradation annual product
    TMF_AC = ee.ImageCollection("projects/JRC/TMF/v1_2024/AnnualChanges").mosaic()
    TMF_AC_2023 = TMF_AC.select("Dec2023")
    Regrowth_TMF = TMF_AC_2023.eq(4)
    return Regrowth_TMF.rename("TMF_regrowth_2023").selfMask()


############tree crops

# TMF_plant (plantations in 2020)
def g_jrc_tmf_plantation_prep():
    transition = ee.ImageCollection(
        "projects/JRC/TMF/v1_2024/TransitionMap_Subtypes"
    ).mosaic()
    deforestation_year = ee.ImageCollection(
        "projects/JRC/TMF/v1_2024/DeforestationYear"
    ).mosaic()
    plantation = (transition.gte(81)).And(transition.lte(86))
    plantation_2020 = plantation.where(
        deforestation_year.gte(2021), 0
    )  # update from https://github.com/forestdatapartnership/whisp/issues/42
    return plantation_2020.rename("TMF_plant").selfMask()


# # Oil_palm_Descals
# NB updated to Descals et al 2024 paper (as opposed to Descals et al 2021 paper)
def g_creaf_descals_palm_prep():
    # Load the Global Oil Palm Year of Plantation image and mosaic it
    img = (
        ee.ImageCollection(
            "projects/ee-globaloilpalm/assets/shared/GlobalOilPalm_YoP_2021"
        )
        .mosaic()
        .select("minNBR_date")
    ).selfMask()

    # Calculate the year of plantation and select all below and including 2020
    oil_palm_plantation_year = img.divide(365).add(1970).floor().lte(2020)

    # Create a mask for plantations in the year 2020 or earlier
    plantation_2020 = oil_palm_plantation_year.lte(2020)
    return plantation_2020.rename("Oil_palm_Descals").selfMask()


# Cocoa_ETH
def g_eth_kalischek_cocoa_prep():
    return (
        ee.Image("projects/ee-nk-cocoa/assets/cocoa_map_threshold_065")
        .rename("Cocoa_ETH")
        .selfMask()
    )


# fdap datasets

# Thresholds and model info here https://github.com/google/forest-data-partnership/blob/main/models/README.md

# Oil Palm FDaP
def g_fdap_palm_prep():
    fdap_palm2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/palm/model_2025a"
    )
    fdap_palm = (
        fdap_palm2020_model_raw.filterDate("2020-01-01", "2020-12-31")
        .mosaic()
        .gt(0.88)  # Precision and recall ~78% at 0.88 threshold.
    )
    return fdap_palm.rename("Oil_palm_FDaP").selfMask()


def g_fdap_palm_2023_prep():
    fdap_palm2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/palm/model_2025a"
    )
    fdap_palm = (
        fdap_palm2020_model_raw.filterDate("2023-01-01", "2023-12-31")
        .mosaic()
        .gt(0.88)  # Precision and recall ~78% at 0.88 threshold.
    )
    return fdap_palm.rename("Oil_palm_2023_FDaP").selfMask()


# Cocoa FDaP
def g_fdap_cocoa_prep():
    fdap_cocoa2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/cocoa/model_2025a"
    )
    fdap_cocoa = (
        fdap_cocoa2020_model_raw.filterDate("2020-01-01", "2020-12-31")
        .mosaic()
        .gt(0.96)  # Precision and recall ~87% 0.96 threshold.
    )
    return fdap_cocoa.rename("Cocoa_FDaP").selfMask()


def g_fdap_cocoa_2023_prep():
    fdap_cocoa2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/cocoa/model_2025a"
    )
    fdap_cocoa = (
        fdap_cocoa2020_model_raw.filterDate("2023-01-01", "2023-12-31")
        .mosaic()
        .gt(0.96)  # Precision and recall ~87% 0.96 threshold.
    )
    return fdap_cocoa.rename("Cocoa_2023_FDaP").selfMask()


# Rubber FDaP
def g_fdap_rubber_prep():
    fdap_rubber2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/rubber/model_2025a"
    )
    fdap_rubber = (
        fdap_rubber2020_model_raw.filterDate("2020-01-01", "2020-12-31")
        .mosaic()
        .gt(0.59)  # Precision and recall ~80% 0.59 threshold.
    )
    return fdap_rubber.rename("Rubber_FDaP").selfMask()


def g_fdap_rubber_2023_prep():
    fdap_rubber2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/rubber/model_2025a"
    )
    fdap_rubber = (
        fdap_rubber2020_model_raw.filterDate("2023-01-01", "2023-12-31")
        .mosaic()
        .gt(0.59)  # Threshold for Rubber
    )
    return fdap_rubber.rename("Rubber_2023_FDaP").selfMask()


# # Coffee FDaP
def g_fdap_coffee_2020_prep():
    # Load the coffee model for 2020
    collection = ee.ImageCollection(
        "projects/forestdatapartnership/assets/coffee/model_2025a"
    )

    # Filter the collection for the year 2020 and create a binary mask
    coffee_2020 = (
        collection.filterDate("2020-01-01", "2020-12-31")
        .mosaic()
        .gt(0.99)  # Precision and recall ~54% 0.99 threshold.
    )

    return coffee_2020.rename("Coffee_FDaP").selfMask()


def g_fdap_coffee_2023_prep():
    # Load the coffee model for 2020
    collection = ee.ImageCollection(
        "projects/forestdatapartnership/assets/coffee/model_2025a"
    )

    # Filter the collection for the year 2023 and create a binary mask
    coffee_2023 = (
        collection.filterDate("2023-01-01", "2023-12-31")
        .mosaic()
        .gt(0.99)  # Precision and recall ~54% 0.99 threshold.
    )
    return coffee_2023.rename("Coffee_FDaP_2023").selfMask()


# Rubber_RBGE  - from Royal Botanical Gardens of Edinburgh (RBGE) NB for 2021
def g_rbge_rubber_prep():
    return (
        ee.Image(
            "users/wangyxtina/MapRubberPaper/rRubber10m202122_perc1585DifESAdist5pxPF"
        )
        .rename("Rubber_RBGE")
        .selfMask()
    )


# soy 2020 South America
def g_soy_song_2020_prep():
    return (
        ee.Image("projects/glad/soy_annual_SA/2020").rename("Soy_Song_2020").selfMask()
    )


##############
# ESRI 2023

# ESRI 2023 - Tree Cover
def g_esri_2023_tc_prep():
    esri_lulc10_raw = ee.ImageCollection(
        "projects/sat-io/open-datasets/landcover/ESRI_Global-LULC_10m_TS"
    )
    esri_lulc10_TC = (
        esri_lulc10_raw.filterDate("2023-01-01", "2023-12-31").mosaic().eq(2)
    )
    return esri_lulc10_TC.rename("ESRI_2023_TC").selfMask()


# ESRI 2023 - Crop
def g_esri_2020_2023_crop_prep():
    esri_lulc10_raw = ee.ImageCollection(
        "projects/sat-io/open-datasets/landcover/ESRI_Global-LULC_10m_TS"
    )
    esri_lulc10_crop_2020 = (
        esri_lulc10_raw.filterDate("2020-01-01", "2020-12-31").mosaic().eq(5)
    )
    esri_lulc10_crop_2023 = (
        esri_lulc10_raw.filterDate("2023-01-01", "2023-12-31").mosaic().eq(5)
    )

    newCrop = esri_lulc10_crop_2023.And(esri_lulc10_crop_2020.Not())

    return newCrop.rename("ESRI_crop_gain_2020_2023").selfMask()


#### disturbances by year

# RADD_year_2019 to RADD_year_< current year >
# Coverage: Primary humid tropical forest areas of South America, sub-Saharan Africa,
#           and insular Southeast Asia at 10m spatial resolution.
#           Available from January 2019 to present for Africa,
#           and from January 2020 to present for South America and Southeast Asia.
def g_radd_year_prep():
    """
    RADD alerts per year as multiband image.
    Each band is binary (1 = alert detected).
    Coverage: Primary humid tropical forest areas of South America, sub-Saharan Africa,
              and insular Southeast Asia at 10m spatial resolution.
              Available from January 2019 to present for Africa,
              and from January 2020 to present for South America and Southeast Asia.
    """
    radd = ee.ImageCollection("projects/radar-wur/raddalert/v1")
    radd_date = (
        radd.filterMetadata("layer", "contains", "alert").select("Date").mosaic()
    )
    start_year = 19
    current_year = CURRENT_YEAR_2DIGIT

    def make_band(year, img_stack):
        start = year * 1000
        end = year * 1000 + 365
        radd_year = (
            radd_date.updateMask(radd_date.gte(start))
            .updateMask(radd_date.lte(end))
            .gt(0)
            .rename("RADD_year_" + "20" + str(year))
            .selfMask()
        )
        return ee.Image(img_stack).addBands(radd_year)

    years = ee.List.sequence(start_year, current_year)
    first_year = ee.Number(years.get(0))
    start = first_year.multiply(1000)
    end = first_year.multiply(1000).add(365)
    band_name = ee.String("RADD_year_").cat("20").cat(first_year.format("%02d"))
    first_band = (
        radd_date.updateMask(radd_date.gte(start))
        .updateMask(radd_date.lte(end))
        .gt(0)
        .rename(band_name)
        .selfMask()
    )

    def make_band(year, img_stack):
        year_num = ee.Number(year)
        start = year_num.multiply(1000)
        end = year_num.multiply(1000).add(365)
        band_name = ee.String("RADD_year_").cat("20").cat(year_num.format("%02d"))
        radd_year = (
            radd_date.updateMask(radd_date.gte(start))
            .updateMask(radd_date.lte(end))
            .gt(0)
            .rename(band_name)
            .selfMask()
        )
        return ee.Image(img_stack).addBands(radd_year)

    img_stack = years.slice(1).iterate(make_band, first_band)
    return ee.Image(img_stack)


# TMF_def_2000 to TMF_def_2023
def g_tmf_def_per_year_prep():
    # Load the TMF Deforestation annual product
    tmf_def = ee.ImageCollection("projects/JRC/TMF/v1_2024/DeforestationYear").mosaic()
    img_stack = None
    # Generate an image based on GFC with one band of forest tree loss per year from 2001 to 2022
    for i in range(0, 24 + 1):
        year_num = ee.Number(2000 + i)
        band_name = ee.String("TMF_def_").cat(year_num.format("%d"))
        tmf_def_year = tmf_def.eq(year_num).rename(band_name).selfMask()
        if img_stack is None:
            img_stack = tmf_def_year
        else:
            img_stack = img_stack.addBands(tmf_def_year)
    return img_stack


# TMF_deg_2000 to TMF_deg_2023
def g_tmf_deg_per_year_prep():
    # Load the TMF Degradation annual product
    tmf_def = ee.ImageCollection("projects/JRC/TMF/v1_2024/DegradationYear").mosaic()
    img_stack = None
    # Generate an image based on GFC with one band of forest tree loss per year from 2001 to 2022
    for i in range(0, 24 + 1):
        year_num = ee.Number(2000 + i)
        band_name = ee.String("TMF_deg_").cat(year_num.format("%d"))
        tmf_def_year = tmf_def.eq(year_num).rename(band_name).selfMask()
        if img_stack is None:
            img_stack = tmf_def_year
        else:
            img_stack = img_stack.addBands(tmf_def_year)
    return img_stack


# GFC_loss_year_2001 to GFC_loss_year_2023 (correct for version 11)
def g_glad_gfc_loss_per_year_prep():
    # Load the Global Forest Change dataset
    gfc = ee.Image("UMD/hansen/global_forest_change_2024_v1_12")
    img_stack = None
    # Generate an image based on GFC with one band of forest tree loss per year from 2001 to 2022
    for i in range(1, 24 + 1):
        year_num = ee.Number(2000 + i)
        band_name = ee.String("GFC_loss_year_").cat(year_num.format("%d"))
        gfc_loss_year = (
            gfc.select(["lossyear"]).eq(i).And(gfc.select(["treecover2000"]).gt(10))
        )
        gfc_loss_year = gfc_loss_year.rename(band_name).selfMask()
        if img_stack is None:
            img_stack = gfc_loss_year
        else:
            img_stack = img_stack.addBands(gfc_loss_year)
    return img_stack


# MODIS_fire_2000 to MODIS_fire_< current year >
def g_modis_fire_prep():
    modis_fire = ee.ImageCollection("MODIS/061/MCD64A1")
    start_year = 2000

    # Determine the last available year by checking the latest image in the collection
    last_image = modis_fire.sort("system:time_start", False).first()
    last_date = ee.Date(last_image.get("system:time_start"))
    end_year = last_date.get("year").getInfo()

    img_stack = None

    for year in range(start_year, end_year + 1):
        year_num = ee.Number(year)
        band_name = ee.String("MODIS_fire_").cat(year_num.format("%d"))
        date_st = f"{year}-01-01"
        date_ed = f"{year}-12-31"
        modis_year = (
            modis_fire.filterDate(date_st, date_ed)
            .mosaic()
            .select(["BurnDate"])
            .gte(0)
            .rename(band_name)
            .selfMask()
        )
        img_stack = modis_year if img_stack is None else img_stack.addBands(modis_year)

    return img_stack


# ESA_fire_2000 to ESA_fire_2020
def g_esa_fire_prep():
    esa_fire = ee.ImageCollection("ESA/CCI/FireCCI/5_1")
    start_year = 2001

    # Determine the last available year by checking the latest image in the collection
    last_image = esa_fire.sort("system:time_start", False).first()
    last_date = ee.Date(last_image.get("system:time_start"))
    end_year = last_date.get("year").getInfo()

    img_stack = None

    for year in range(start_year, end_year + 1):
        year_num = ee.Number(year)
        band_name = ee.String("ESA_fire_").cat(year_num.format("%d"))
        date_st = f"{year}-01-01"
        date_ed = f"{year}-12-31"
        esa_year = (
            esa_fire.filterDate(date_st, date_ed)
            .mosaic()
            .select(["BurnDate"])
            .gte(0)
            .rename(band_name)
            .selfMask()
        )
        img_stack = esa_year if img_stack is None else img_stack.addBands(esa_year)

    return img_stack


#### disturbances combined (split into before and after 2020)

# RADD_after_2020
def g_radd_after_2020_prep():
    radd = ee.ImageCollection("projects/radar-wur/raddalert/v1")

    radd_date = (
        radd.filterMetadata("layer", "contains", "alert").select("Date").mosaic()
    )
    # date of avaialbility
    start_year = 21  ## (starts 2019 in Africa, then 2020 for S America and Asia: https://data.globalforestwatch.org/datasets/gfw::deforestation-alerts-radd/about)

    # Use pre-calculated current year (avoids repeated datetime calls)
    current_year = CURRENT_YEAR_2DIGIT
    start = start_year * 1000
    end = current_year * 1000 + 365
    return (
        radd_date.updateMask(radd_date.gte(start))
        .updateMask(radd_date.lte(end))
        .gt(0)
        .rename("RADD_after_2020")
    ).selfMask()


# RADD_before_2020
def g_radd_before_2020_prep():
    radd = ee.ImageCollection("projects/radar-wur/raddalert/v1")

    radd_date = (
        radd.filterMetadata("layer", "contains", "alert").select("Date").mosaic()
    )
    # date of avaialbility
    start_year = 19  ## (starts 2019 in Africa, then 2020 for S America and Asia: https://data.globalforestwatch.org/datasets/gfw::deforestation-alerts-radd/about)

    start = start_year * 1000
    end = 20 * 1000 + 365
    return (
        radd_date.updateMask(radd_date.gte(start))
        .updateMask(radd_date.lte(end))
        .gt(0)
        .rename("RADD_before_2020")
    ).selfMask()


# DIST_after_2020

# DIST alerts are for all veg types so masked by EUFO forest 2020
# NB alerts only for 2024 onwards (in GEE at least, available for 2023 ofrom the GLAD site)
# for conistency using "...after_2020..." terminology.
def g_glad_dist_after_2020_prep():

    # no need to filter by date as all dates are later than 2023

    # Load the vegetation disturbance collections
    VEGDISTSTATUS = ee.ImageCollection(
        "projects/glad/HLSDIST/current/VEG-DIST-STATUS"
    ).mosaic()

    # Key for high-confidence alerts (values 3, 6, 7, 8)
    high_conf_values = [3, 6, 7, 8]

    # Create high-confidence mask
    dist_high_conf = VEGDISTSTATUS.remap(
        high_conf_values, [1] * len(high_conf_values), 0
    )

    return dist_high_conf.updateMask(g_jrc_gfc_2020_prep()).rename(
        "DIST_after_2020"
    )  # Mask alerts to forest and rename band


# # DIST_alert_2024 to DIST_alert_< current year >
# # Notes:
# # 1) so far only available for 2024 onwards in GEE
# # 2) masked alerts (as dist alerts are for all vegetation) to JRC EUFO 2020 layer, as close to EUDR definition


def g_glad_dist_year_prep():
    """
    GLAD DIST alerts per year as multiband image.
    Each band is binary (1 = high-confidence disturbance alert).
    Uses VEG-DIST-DATE to filter by year, VEG-DIST-STATUS for confidence.
    Masked to EUFO 2020 forest.
    Note: Only available from 2024 onwards.
    Fully server-side using ee.List.iterate (no Python for loop).
    """
    # Load the vegetation disturbance collections
    #  Vegetation disturbance status (0-8, class flag, 8-bit)
    VEGDISTSTATUS = ee.ImageCollection(
        "projects/glad/HLSDIST/current/VEG-DIST-STATUS"
    ).mosaic()
    # Initial vegetation disturbance date (>0: days since 2020-12-31, 16-bit)
    VEGDISTDATE = ee.ImageCollection(
        "projects/glad/HLSDIST/current/VEG-DIST-DATE"
    ).mosaic()

    # Key for high-confidence alerts (values 3, 6, 7, 8)
    # 3 = <50% loss, high confidence, ongoing
    # 6 = ≥50% loss, high confidence, ongoing
    # 7 = <50% loss, high confidence, finished
    # 8 = ≥50% loss, high confidence, finished
    high_conf_values = [3, 6, 7, 8]
    dist_high_conf = VEGDISTSTATUS.remap(
        high_conf_values, [1] * len(high_conf_values), 0
    )

    # Year range: 2024 to current year
    start_year = 2024
    end_year = CURRENT_YEAR

    # Reference date for day offset calculation (2020-12-31)
    ref_date = ee.Date("2020-12-31")

    # Create first band (2024)
    first_year = ee.Number(start_year)
    first_start_days = ee.Date.fromYMD(first_year, 1, 1).difference(ref_date, "day")
    first_end_days = ee.Date.fromYMD(first_year.add(1), 1, 1).difference(
        ref_date, "day"
    )
    first_year_mask = VEGDISTDATE.gte(first_start_days).And(
        VEGDISTDATE.lt(first_end_days)
    )
    first_band_name = ee.String("DIST_year_").cat(first_year.format("%d"))
    first_band = (
        first_year_mask.updateMask(dist_high_conf).rename(first_band_name).selfMask()
    )

    # Server-side iteration to add remaining years
    years = ee.List.sequence(start_year + 1, end_year)

    def add_year_band(year, img_stack):
        year_num = ee.Number(year)
        start_days = ee.Date.fromYMD(year_num, 1, 1).difference(ref_date, "day")
        end_days = ee.Date.fromYMD(year_num.add(1), 1, 1).difference(ref_date, "day")
        year_mask = VEGDISTDATE.gte(start_days).And(VEGDISTDATE.lt(end_days))
        band_name = ee.String("DIST_year_").cat(year_num.format("%d"))
        year_band = year_mask.updateMask(dist_high_conf).rename(band_name).selfMask()
        return ee.Image(img_stack).addBands(year_band)

    img_stack = ee.Image(years.iterate(add_year_band, first_band))

    # Mask to EUFO 2020 forest
    return img_stack.updateMask(g_jrc_gfc_2020_prep())


# GLAD-L (GLAD Landsat) Alerts
# Coverage: Entire tropics (30°N to 30°S) from January 1, 2018 to present,
#           and from 2015 to present for select countries in the Amazon, Congo Basin,
#           and insular Southeast Asia.
# Uses confidence bands per year where values >= 2 are confirmed alerts
# Asset paths per year:
#   2021: projects/glad/alert/2021final (conf21)
#   2022: projects/glad/alert/2022final (conf22)
#   2023: projects/glad/alert/2023final (conf23)
#   2024: NOT AVAILABLE
#   2025+: projects/glad/alert/UpdResult (conf25, conf26, etc.)
# More info: https://glad.umd.edu/dataset/glad-forest-alerts


# GLAD-L_after_2020 (combined alerts from 2021 to current year, excluding 2024)
def g_glad_l_after_2020_prep():
    """
    GLAD Landsat alerts after 2020 (combined from 2021 onwards).
    Uses confidence bands with threshold >= 2 for confirmed alerts.
    Note: 2024 data is not available.
    """
    # Load yearly assets and combine into single multiband image
    glad_combined = (
        ee.ImageCollection("projects/glad/alert/2021final")
        .mosaic()
        .select("conf21")
        .addBands(
            ee.ImageCollection("projects/glad/alert/2022final")
            .mosaic()
            .select("conf22")
        )
        .addBands(
            ee.ImageCollection("projects/glad/alert/2023final")
            .mosaic()
            .select("conf23")
        )
        .addBands(
            ee.ImageCollection("projects/glad/alert/UpdResult")
            .mosaic()
            .select(["conf25", "conf26"])
        )
    )

    # Combine alerts from all available years (confidence >= 2)
    # 2024 not available
    combined_alerts = (
        glad_combined.select("conf21")
        .gte(2)
        .Or(glad_combined.select("conf22").gte(2))
        .Or(glad_combined.select("conf23").gte(2))
        .Or(glad_combined.select("conf25").gte(2))
        .Or(glad_combined.select("conf26").gte(2))
    )

    return combined_alerts.rename("GLAD-L_after_2020").selfMask()


# GLAD-L_before_2020 (combined alerts from 2017 to 2020)
def g_glad_l_before_2020_prep():
    """
    GLAD Landsat alerts before 2020 (combined from 2017-2020 inclusive).
    Uses confidence bands with threshold >= 2 for confirmed alerts.
    Note: 2015 and 2016 assets are not available in GEE.
    Coverage: Tropics (30°N to 30°S).
    """
    # Load yearly assets and combine
    glad_combined = (
        ee.ImageCollection("projects/glad/alert/2017final")
        .mosaic()
        .select("conf17")
        .addBands(
            ee.ImageCollection("projects/glad/alert/2018final")
            .mosaic()
            .select("conf18")
        )
        .addBands(
            ee.ImageCollection("projects/glad/alert/2019final")
            .mosaic()
            .select("conf19")
        )
        .addBands(
            ee.ImageCollection("projects/glad/alert/2020final")
            .mosaic()
            .select("conf20")
        )
    )

    # Combine alerts from all available years (confidence >= 2)
    combined_alerts = (
        glad_combined.select("conf17")
        .gte(2)
        .Or(glad_combined.select("conf18").gte(2))
        .Or(glad_combined.select("conf19").gte(2))
        .Or(glad_combined.select("conf20").gte(2))
    )

    return combined_alerts.rename("GLAD-L_before_2020").selfMask()


# GLAD-L timeseries - multiband image with one band per year
def g_glad_l_year_prep():
    """
    GLAD Landsat alerts per year as multiband image.
    Each band is binary (1 = alert with confidence >= 2).
    Coverage: Entire tropics (30°N to 30°S) from January 1, 2018 to present,
              and from 2017 to present for select countries in the Amazon,
              Congo Basin, and insular Southeast Asia.
    Note: 2015 and 2016 assets are not available in GEE.
    Note: 2024 data is not available.
    Includes years from 2017 onwards.
    """
    # Build multiband image with all available years
    # Years 2017-2023 use YYYYfinal assets, 2025+ use UpdResult
    # Note: 2015final and 2016final assets do not exist in GEE
    img_stack = (
        # 2017
        ee.ImageCollection("projects/glad/alert/2017final")
        .mosaic()
        .select("conf17")
        .gte(2)
        .rename("GLAD-L_year_2017")
        .selfMask()
        # 2018
        .addBands(
            ee.ImageCollection("projects/glad/alert/2018final")
            .mosaic()
            .select("conf18")
            .gte(2)
            .rename("GLAD-L_year_2018")
            .selfMask()
        )
        # 2019
        .addBands(
            ee.ImageCollection("projects/glad/alert/2019final")
            .mosaic()
            .select("conf19")
            .gte(2)
            .rename("GLAD-L_year_2019")
            .selfMask()
        )
        # 2020
        .addBands(
            ee.ImageCollection("projects/glad/alert/2020final")
            .mosaic()
            .select("conf20")
            .gte(2)
            .rename("GLAD-L_year_2020")
            .selfMask()
        )
        # 2021
        .addBands(
            ee.ImageCollection("projects/glad/alert/2021final")
            .mosaic()
            .select("conf21")
            .gte(2)
            .rename("GLAD-L_year_2021")
            .selfMask()
        )
        # 2022
        .addBands(
            ee.ImageCollection("projects/glad/alert/2022final")
            .mosaic()
            .select("conf22")
            .gte(2)
            .rename("GLAD-L_year_2022")
            .selfMask()
        )
        # 2023
        .addBands(
            ee.ImageCollection("projects/glad/alert/2023final")
            .mosaic()
            .select("conf23")
            .gte(2)
            .rename("GLAD-L_year_2023")
            .selfMask()
        )
        # 2024 NOT AVAILABLE
        # 2025
        .addBands(
            ee.ImageCollection("projects/glad/alert/UpdResult")
            .mosaic()
            .select("conf25")
            .gte(2)
            .rename("GLAD-L_year_2025")
            .selfMask()
        )
        # 2026
        .addBands(
            ee.ImageCollection("projects/glad/alert/UpdResult")
            .mosaic()
            .select("conf26")
            .gte(2)
            .rename("GLAD-L_year_2026")
            .selfMask()
        )
    )

    return img_stack


# GLAD-S2 (GLAD Sentinel-2) Alerts
# GLAD-S2_after_2020 (combined alerts from 2021 to current year)
def g_glad_s2_after_2020_prep():
    """
    GLAD Sentinel-2 alerts after 2020 (filtered and combined from 2021 onwards - original data starts 2019).
    Uses alert band with threshold >= 2 for confirmed alerts.
    Coverage: Primary humid tropical forest within Amazon basin region.
    https://glad.umd.edu/dataset/glad-forest-alerts
    """
    col = "projects/glad/S2alert"
    s2alert = ee.Image(col + "/alert")
    alert_date = ee.Image(col + "/alertDate")

    # Date encoding: days since 2019-01-01
    # 2020-12-31 = 730 days (end of 2020)
    days_end_2020 = 730

    # Filter alerts after 2020 with confidence >= 2
    #   - alert: confidence band (0-4, where 4 is highest confidence)
    alerts_after_2020 = s2alert.gte(2).And(alert_date.gt(days_end_2020))

    return alerts_after_2020.rename("GLAD-S2_after_2020").selfMask()


# GLAD-S2_before_2020 (combined alerts from 2019 to 2020)
def g_glad_s2_before_2020_prep():
    """
    GLAD Sentinel-2 alerts before 2020 (from 2019-01-01 to 2020-12-31 inclusive).
    Uses alert band with threshold >= 2 for confirmed alerts.
    Coverage: Primary humid tropical forest within Amazon basin region.
    Note: Data starts from 2019.
    """
    col = "projects/glad/S2alert"
    s2alert = ee.Image(col + "/alert")
    alert_date = ee.Image(col + "/alertDate")

    # Date encoding: days since 2019-01-01
    # 2019-01-01 = 0, 2020-12-31 = 730 days
    days_end_2020 = 730

    # Filter alerts up to end of 2020 with confidence >= 2
    alerts_before_2020 = s2alert.gte(2).And(alert_date.lte(days_end_2020))

    return alerts_before_2020.rename("GLAD-S2_before_2020").selfMask()


# GLAD-S2 timeseries - multiband image with one band per year
def g_glad_s2_year_prep():
    """
    GLAD Sentinel-2 alerts per year as multiband image.
    Each band is binary (1 = alert with confidence >= 2).
    Coverage: Primary humid tropical forest areas of South America
              from January 2019 to present.
    Date encoding: days since 2019-01-01.
    Includes years from 2019 onwards.
    """
    col = "projects/glad/S2alert"
    s2alert = ee.Image(col + "/alert")
    alert_date = ee.Image(col + "/alertDate")

    # Confidence threshold for confirmed alerts
    confirmed = s2alert.gte(2)

    # Date encoding: days since 2019-01-01
    # Year boundaries (days since 2019-01-01):
    # 2019-01-01 = 0, 2020-01-01 = 365, 2021-01-01 = 731, 2022-01-01 = 1096
    # 2023-01-01 = 1461, 2024-01-01 = 1827, 2025-01-01 = 2192, 2026-01-01 = 2557
    ref_date = ee.Date("2019-01-01")

    # Build multiband image with available years (data starts 2019)
    start_year = 2019
    end_year = CURRENT_YEAR

    # Create first band (2019)
    first_year = ee.Number(start_year)
    first_start_days = ee.Date.fromYMD(first_year, 1, 1).difference(ref_date, "day")
    first_end_days = ee.Date.fromYMD(first_year.add(1), 1, 1).difference(
        ref_date, "day"
    )
    first_year_mask = alert_date.gte(first_start_days).And(
        alert_date.lt(first_end_days)
    )
    first_band_name = ee.String("GLAD-S2_year_").cat(first_year.format("%d"))
    first_band = (
        confirmed.updateMask(first_year_mask).rename(first_band_name).selfMask()
    )

    # Server-side iteration to add remaining years
    years = ee.List.sequence(start_year + 1, end_year)

    def add_year_band(year, img_stack):
        year_num = ee.Number(year)
        start_days = ee.Date.fromYMD(year_num, 1, 1).difference(ref_date, "day")
        end_days = ee.Date.fromYMD(year_num.add(1), 1, 1).difference(ref_date, "day")
        year_mask = alert_date.gte(start_days).And(alert_date.lt(end_days))
        band_name = ee.String("GLAD-S2_year_").cat(year_num.format("%d"))
        year_band = confirmed.updateMask(year_mask).rename(band_name).selfMask()
        return ee.Image(img_stack).addBands(year_band)

    img_stack = ee.Image(years.iterate(add_year_band, first_band))

    return img_stack


#### disturbances combined (split into before and after 2020)

# TMF_deg_before_2020
def g_tmf_deg_before_2020_prep():
    tmf_deg = ee.ImageCollection("projects/JRC/TMF/v1_2024/DegradationYear").mosaic()
    return (
        (tmf_deg.lte(2020))
        .And(tmf_deg.gte(2000))
        .rename("TMF_deg_before_2020")
        .selfMask()
    )


# TMF_deg_after_2020
def g_tmf_deg_after_2020_prep():
    tmf_deg = ee.ImageCollection("projects/JRC/TMF/v1_2024/DegradationYear").mosaic()
    return tmf_deg.gt(2020).rename("TMF_deg_after_2020").selfMask()


# tmf_def_before_2020
def g_tmf_def_before_2020_prep():
    tmf_def = ee.ImageCollection("projects/JRC/TMF/v1_2024/DeforestationYear").mosaic()
    return (
        (tmf_def.lte(2020))
        .And(tmf_def.gte(2000))
        .rename("TMF_def_before_2020")
        .selfMask()
    )


# tmf_def_after_2020
def g_tmf_def_after_2020_prep():
    tmf_def = ee.ImageCollection("projects/JRC/TMF/v1_2024/DeforestationYear").mosaic()
    return tmf_def.gt(2020).rename("TMF_def_after_2020").selfMask()


# GFC_loss_before_2020 (loss within 10 percent cover; includes 2020; correct for version 11)
def g_glad_gfc_loss_before_2020_prep():
    # Load the Global Forest Change dataset
    gfc = ee.Image("UMD/hansen/global_forest_change_2024_v1_12")
    gfc_loss = (
        gfc.select(["lossyear"]).lte(20).And(gfc.select(["treecover2000"]).gt(10))
    )
    return gfc_loss.rename("GFC_loss_before_2020").selfMask()


# GFC_loss_after_2020 (loss within 10 percent cover; correct for version 11)
def g_glad_gfc_loss_after_2020_prep():
    # Load the Global Forest Change dataset
    gfc = ee.Image("UMD/hansen/global_forest_change_2024_v1_12")
    gfc_loss = gfc.select(["lossyear"]).gt(20).And(gfc.select(["treecover2000"]).gt(10))
    return gfc_loss.rename("GFC_loss_after_2020").selfMask()


# MODIS_fire_before_2020
def g_modis_fire_before_2020_prep():
    modis_fire = ee.ImageCollection("MODIS/061/MCD64A1")
    start_year = 2000
    end_year = 2020
    date_st = str(start_year) + "-01-01"
    date_ed = str(end_year) + "-12-31"
    return (
        modis_fire.filterDate(date_st, date_ed)
        .mosaic()
        .select(["BurnDate"])
        .gte(0)
        .rename("MODIS_fire_before_2020")
    ).selfMask()


# MODIS_fire_after_2020
def g_modis_fire_after_2020_prep():
    modis_fire = ee.ImageCollection("MODIS/061/MCD64A1")
    start_year = 2021
    # Use pre-calculated current year (avoids repeated datetime calls)
    end_year = CURRENT_YEAR - 1  # Use year - 1 to ensure data availability
    date_st = str(start_year) + "-01-01"
    date_ed = str(end_year) + "-12-31"
    return (
        modis_fire.filterDate(date_st, date_ed)
        .mosaic()
        .select(["BurnDate"])
        .gte(0)
        .rename("MODIS_fire_after_2020")
    ).selfMask()


# ESA_fire_before_2020
def g_esa_fire_before_2020_prep():
    esa_fire = ee.ImageCollection("ESA/CCI/FireCCI/5_1")
    start_year = 2000
    end_year = 2020
    date_st = str(start_year) + "-01-01"
    date_ed = str(end_year) + "-12-31"
    return (
        esa_fire.filterDate(date_st, date_ed)
        .mosaic()
        .select(["BurnDate"])
        .gte(0)
        .rename("ESA_fire_before_2020")
    ).selfMask()


#########################logging concessions
# http://data.globalforestwatch.org/datasets?q=logging&sort_by=relevance
def g_logging_concessions_before_2020_prep():
    RCA = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/RCA_Permis_dExploitation_et_dAmenagement"
    )
    RCA_binary = ee.Image().paint(RCA, 1)
    CMR = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/Cameroon_Forest_Management_Units"
    )
    CMR_binary = ee.Image().paint(CMR, 1)
    Eq_G = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/Equatorial_Guinea_logging_concessions"
    )
    Eq_G_binary = ee.Image().paint(Eq_G, 1)
    DRC = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/DRC_Forest_concession_agreements"
    )
    DRC_binary = ee.Image().paint(DRC, 1)
    Liberia = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/Liberia_Forest_Management_Contracts"
    )
    Liberia_binary = ee.Image().paint(Liberia, 1)
    RoC = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/Republic_of_the_Congo_logging_concessions"
    )
    Roc_binary = ee.Image().paint(RoC, 1)
    Sarawak = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/Sarawak_logging_concessions"
    )
    Sarawak_binary = ee.Image().paint(Sarawak, 1)
    logging_concessions_binary = ee.ImageCollection(
        [
            RCA_binary,
            CMR_binary,
            Eq_G_binary,
            DRC_binary,
            Liberia_binary,
            Roc_binary,
            Sarawak_binary,
        ]
    ).mosaic()

    return logging_concessions_binary.rename("GFW_logging_before_2020").selfMask()


#########################national datasets

# nBR Brazil

# ### nBR Natural forests in 2020:

# %%
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Amazon biome, 2020
# Subsetting criteria: primary forests (DN=1) and secondary forests (DN=2) // secondary forests are those recovering from deforestation
# the resulting dataset shows primary and secondary forest cover in 2020 (mostly by August 2020)

##########################primary forests###############################################
def nbr_terraclass_amz20_primary_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_f = tcamz20.eq(1)
    return tcamz20_f.rename("nBR_INPE_TC_primary_forest_Amazon_2020").selfMask()


# [Official NFMS dataset] Brazilian Forest Service dataset on natural forest cover from PRODES and TerraClass data, base year 2022
# Subsetting criteria: ano_desmat > 2020 and nom_class = 'Floresta'
# the resulting datasets show primary forest cover in 2020 for the Pantanal, Caatinga, Atlantic Forest and Pampa biomes.
# the resulting dataset shows primary and secondary forest cover in 2020 for the Cerrado biome (TerraClass 2020)
# For the Amazon, best to use Terraclass 2020 directly, because the BFS used TerraClass 2014.

# Pantanal
def nbr_bfs_ptn_f20_prep():
    bfs_fptn20 = ee.FeatureCollection("projects/ee-whisp/assets/NBR/bfs_ptn_2020")

    bfs_fptn20_binary = ee.Image().paint(bfs_fptn20, 1)
    return bfs_fptn20_binary.rename("nBR_BFS_primary_forest_Pantanal_2020").selfMask()


# Caatinga - filtered with QGIS because the original geodatabase is too large to export as a shapefile (GEE accepted format)
## couldn't convert it to asset, working on it (Error: Primary geometry of feature '306862' has 2454627 vertices, above the limit of 1000000 vertices. (Error code: 3)
def nbr_bfs_caat_f20_prep():
    bfs_fcaat20 = ee.FeatureCollection("projects/ee-whisp/assets/NBR/bfs_caat_2020")
    bfs_fcaat20_binary = ee.Image().paint(bfs_fcaat20, 1)
    return bfs_fcaat20_binary.rename("nBR_BFS_primary_forest_Caatinga_2020").selfMask()


# Atlantic Forest - filtered with QGIS because the original geodatabase is too large to export as a shapefile (GEE accepted format)
def nbr_bfs_atlf_f20_prep():
    bfs_fatlf20 = ee.FeatureCollection("projects/ee-whisp/assets/NBR/bfs_atlf_2020")
    bfs_fatlf20_binary = ee.Image().paint(bfs_fatlf20, 1)
    return bfs_fatlf20_binary.rename(
        "nBR_BFS_primary_forest_AtlanticForest_2020"
    ).selfMask()


# Pampa - filtered in QGIS to save some storage space
def nbr_bfs_pmp_f20_prep():
    bfs_fpmp20 = ee.FeatureCollection("projects/ee-whisp/assets/NBR/bfs_pmp_2020")
    bfs_fpmp20_binary = ee.Image().paint(bfs_fpmp20, 1)
    return bfs_fpmp20_binary.rename("nBR_BFS_primary_forest_Pampa_2020").selfMask()


##########################secondary forests###############################################
def nbr_terraclass_amz20_secondary_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_f = tcamz20.eq(2)
    return tcamz20_f.rename("nBR_INPE_TC_secondary_forest_Amazon_2020").selfMask()


# Cerrado - filtered with QGIS because the original geodatabase is too large to export as a shapefile (GEE accepted format)
def nbr_bfs_cer_f20_prep():
    bfs_fcer20 = ee.FeatureCollection("projects/ee-whisp/assets/NBR/bfs_cerr_2020")
    bfs_fcer20_binary = ee.Image().paint(bfs_fcer20, 1)
    return bfs_fcer20_binary.rename(
        "nBR_BFS_primary_and_secondary_forest_Cerrado_2020"
    ).selfMask()


# %%
# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: classification_2020 = Forest formation (DN=3), Savanna Formation (DN=4, forest according to BR definition), Mangrove (DN=5), Floodable Forest (DN=6), Wooded Sandbank veg (DN=49)
# the resulting dataset shows forest cover in 2020, without distinguishing between primary and secondary forests
def nbr_mapbiomasc9_f20_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_forest = (
        mapbiomasc9_20.eq(3)
        .Or(mapbiomasc9_20.eq(4))
        .Or(mapbiomasc9_20.eq(5))
        .Or(mapbiomasc9_20.eq(6))
        .Or(mapbiomasc9_20.eq(49))
    )
    return mapbiomasc9_20_forest.rename(
        "nBR_MapBiomas_col9_forest_Brazil_2020"
    ).selfMask()


# ### ########################NBR plantation forest in 2020:#######################################

# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Amazon biome, 2020
# Subsetting criteria: silviculture (DN=9)
# the resulting dataset shows monospecific commercial plantations, mostly eucalyptus and pinus.
def nbr_terraclass_amz20_silv_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_silviculture = tcamz20.eq(9)
    return tcamz20_silviculture.rename("nBR_INPE_TCsilviculture_Amazon_2020").selfMask()


# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Cerrado biome, 2020
# Subsetting criteria: silviculture (DN=9)
# the resulting dataset shows monospecific commercial plantations, mostly eucalyptus and pinus.
def nbr_terraclass_silv_cer20_prep():
    tccer20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_cer_2020")
    tccer20_silviculture = tccer20.eq(9)
    return tccer20_silviculture.rename(
        "nBR_INPE_TCsilviculture_Cerrado_2020"
    ).selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = Forest plantation (DN=9)
# the resulting dataset shows forest plantation in 2020
def nbr_mapbiomasc9_silv20_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_silviculture = mapbiomasc9_20.eq(9)
    return mapbiomasc9_20_silviculture.rename(
        "nBR_MapBiomas_col9_silviculture_Brazil_2020"
    ).selfMask()


################ ### NBR Disturbances before 2020:########################################

# [Official NFMS dataset] INPE PRODES data up to 2023
# Subsetting criteria: DN = [0, 2, 4, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60];

# the resulting dataset shows deforestation and conversion of OWL and OL up to 2020 (mostly August 2020), including residues (omission errors corrections)
def nbr_prodes_before_2020_prep():
    prodes = ee.Image("projects/ee-whisp/assets/NBR/prodes_brasil_2023")
    prodes_before_20_dn = [
        0,
        2,
        4,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        50,
        51,
        52,
        53,
        54,
        55,
        56,
        57,
        58,
        59,
        60,
    ]
    prodes_before_20_mask = prodes.remap(
        prodes_before_20_dn, [1] * len(prodes_before_20_dn)
    )  # .eq(1)
    return prodes_before_20_mask.rename(
        "nBR_PRODES_deforestation_Brazil_before_2020"
    ).selfMask()


## Caution: 1) includes deforestation and conversion of other wooded land and grassland

# [Official NFMS dataset] INPE.DETER data from 2nd August 2016 up to the 04th of April 2025
# Subsetting criteria: forest degradation classes ['CICATRIZ_DE_QUEIMADA', 'CS_DESORDENADO', 'DEGRADACAO'] and view_date until 2020-12-31
# 'CS_GEOMETRICO' excluded to align with FREL


def nbr_deter_amazon_before_2020_prep():
    deteramz = ee.FeatureCollection("projects/ee-whisp/assets/NBR/deter_amz_16apr2025")
    degradation_classes = ["CICATRIZ_DE_QUEIMADA", "CS_DESORDENADO", "DEGRADACAO"]

    # Add a formatted date field based on VIEW_DATE
    def add_formatted_date(feature):
        return feature.set("formatted_date", ee.Date(feature.get("VIEW_DATE")))

    deteramz = deteramz.map(add_formatted_date)

    deter_deg = deteramz.filter(
        ee.Filter.inList("CLASSNAME", degradation_classes)
    ).filter(ee.Filter.lt("formatted_date", ee.Date("2020-12-31")))

    deter_deg_binary = ee.Image().paint(deter_deg, 1)
    return deter_deg_binary.rename(
        "nBR_DETER_forestdegradation_Amazon_before_2020"
    ).selfMask()


################ ### NBR Disturbances after 2020:########################################
# [Official NFMS dataset] INPE PRODES data up to 2023
# Subsetting criteria: DN = [21, 22, 23, 61, 62, 63];

# the resulting dataset shows deforestation and conversion of OWL and OL up to 2020 (mostly August 2020), including residues (omission errors corrections)


def nbr_prodes_after_2020_prep():
    prodes = ee.Image("projects/ee-whisp/assets/NBR/prodes_brasil_2023")
    prodes_after_20_dn = [21, 22, 23, 61, 62, 63]
    prodes_after_20_mask = prodes.remap(
        prodes_after_20_dn, [1] * len(prodes_after_20_dn)
    )  # .eq(1)
    prodes_after_20 = prodes_after_20_mask.selfMask()
    return prodes_after_20.rename(
        "nBR_PRODES_deforestation_Brazil_after_2020"
    ).selfMask()


# %%
# [Official NFMS dataset] INPE.DETER data from 2nd August 2016 up to the 04th of April 2025
# Subsetting criteria: forest degradation classes ['CICATRIZ_DE_QUEIMADA', 'CS_DESORDENADO', 'DEGRADACAO'] and view_date from 2021-01-01 onward
# 'CS_GEOMETRICO' excluded to align with FREL
def nbr_deter_amazon_after_2020_prep():
    deteramz = ee.FeatureCollection("projects/ee-whisp/assets/NBR/deter_amz_16apr2025")
    degradation_classes = ["CICATRIZ_DE_QUEIMADA", "CS_DESORDENADO", "DEGRADACAO"]

    # Add a formatted date field based on VIEW_DATE
    def add_formatted_date(feature):
        return feature.set("formatted_date", ee.Date(feature.get("VIEW_DATE")))

    deteramz = deteramz.map(add_formatted_date)

    deter_deg = deteramz.filter(
        ee.Filter.inList("CLASSNAME", degradation_classes)
    ).filter(ee.Filter.gt("formatted_date", ee.Date("2021-01-01")))

    deter_deg_binary = ee.Image().paint(deter_deg, 1)
    return deter_deg_binary.rename(
        "nBR_DETER_forestdegradation_Amazon_after_2020"
    ).selfMask()


# ########################## NBR commodities - permanent/perennial crops in 2020:###############################
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Amazon biome, 2020
# OR [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Cerrado biome, 2020
# Subsetting criteria: perennial (DN=12) and semi-perennial (DN=13) crops
# the resulting dataset shows perennial and semi-perennial crops in 2020
def nbr_terraclass_amz_cer20_pc_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_pc = tcamz20.eq(12).Or(tcamz20.eq(13))
    tccer20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_cer_2020")
    tccer20_pc = tccer20.eq(12).Or(tccer20.eq(13))
    tc_pc = ee.ImageCollection([tcamz20_pc, tccer20_pc]).mosaic()
    return tc_pc.rename("nBR_INPE_TCamz_cer_perennial_2020").selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = coffee (DN=46) <================== COFFEE
# the resulting dataset shows coffee area in 2020
def nbr_mapbiomasc9_cof_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_coffee = mapbiomasc9_20.eq(46)
    return mapbiomasc9_20_coffee.rename("nBR_MapBiomas_col9_coffee_2020").selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = palm oil (DN=35) <================= PALM OIL
# the resulting dataset shows palm oil area in 2020
def nbr_mapbiomasc9_po_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_palm = mapbiomasc9_20.eq(35)
    return mapbiomasc9_20_palm.rename("nBR_MapBiomas_col9_palmoil_2020").selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = other perennial crops (DN=48)
# the resulting dataset shows citrus and perennial crops other than coffee and palm oil in 2020
def nbr_mapbiomasc9_pc_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_pc = mapbiomasc9_20.eq(35).Or(mapbiomasc9_20.eq(46))
    return mapbiomasc9_20_pc.rename("nBR_MapBiomas_col9_pc_2020").selfMask()


# ######################## NBR commodities - annual crops in 2020:##############################

# %%
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Amazon biome, 2020
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Cerrado biome, 2020
# Subsetting criteria: annual/temporary 1 cycle (DN=14) or more than 1 cycle (DN=15)
# the resulting dataset shows temporary crop in 2020
def nbr_terraclass_amz_cer20_ac_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_ac = tcamz20.eq(14).Or(tcamz20.eq(15))
    tccer20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_cer_2020")
    tccer20_ac = tccer20.eq(14).Or(tccer20.eq(15))
    tc_ac = ee.ImageCollection([tcamz20_ac, tccer20_ac]).mosaic()
    return tc_ac.rename("nBR_INPE_TCamz_cer_annual_2020").selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = soybean (DN=39) <================== SOY
# the resulting dataset shows soybean plantation area in 2020
def nbr_mapbiomasc9_soy_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_soy = mapbiomasc9_20.eq(39)
    return mapbiomasc9_20_soy.rename("nBR_MapBiomas_col9_soy_2020").selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = other temporary crops (DN=41)
# Subsetting criteria: 'classification_2020' = sugar cane (DN=20)
# Subsetting criteria: 'classification_2020' = rice (DN=40)
# Subsetting criteria: 'classification_2020' = cotton (beta version, DN=62)
# the resulting dataset shows temporary crop area other than soy, includes sugar cane, rice, and cotton
def nbr_mapbiomasc9_ac_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_ac = (
        mapbiomasc9_20.eq(41)
        .Or(mapbiomasc9_20.eq(20))
        .Or(mapbiomasc9_20.eq(40))
        .Or(mapbiomasc9_20.eq(62))
    )
    return mapbiomasc9_20_ac.rename("nBR_MapBiomas_col9_annual_crops_2020").selfMask()


# ################################### NBR commodities - pasture/livestock in 2020:##############################

# %%
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Amazon biome, 2020
# Subsetting criteria: BUSH/SHRUB PASTURE (DN=10) or HERBACEOUS PASTURE (DN=11)

# the resulting dataset shows 2020 pasture area in the Amazon
def nbr_terraclass_amz20_pasture_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_pasture = tcamz20.eq(10).Or(tcamz20.eq(11))
    return tcamz20_pasture.rename("nBR_INPE_TCamz_pasture_2020").selfMask()


# %%
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Cerrado biome, 2020
# Subsetting criteria: PASTURE (DN=11)
# the resulting dataset shows 2020 pasture area in the Cerrado


def nbr_terraclass_cer20_ac_prep():
    tccer20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_cer_2020")
    tccer20_pasture = tccer20.eq(11)
    return tccer20_pasture.rename("nBR_INPE_TCcer_pasture_2020").selfMask()


# %%
# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = pasture (DN=15)
# the resulting dataset shows pasture area in 2020 in Brazil
def nbr_mapbiomasc9_pasture_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_pasture = mapbiomasc9_20.eq(15)
    return mapbiomasc9_20_pasture.rename("nBR_MapBiomas_col9_pasture_2020").selfMask()


###################################################################
# nCO - Colombia


def nco_ideam_forest_2020_prep():
    ideam_forest_raw = ee.Image("projects/ee-whisp/assets/nCO/ideam_2020_geo")
    ideam_forest = ideam_forest_raw.eq(1)  # get forest class
    return ideam_forest.rename("nCO_ideam_forest_2020").selfMask()


def nco_ideam_eufo_commission_2020_prep():
    ideam_agroforest_raw = ee.Image("projects/ee-whisp/assets/nCO/ideam_2020_geo_EUFO")
    ideam_agroforest = ideam_agroforest_raw.eq(4)  # get forest class
    return ideam_agroforest.rename("nCO_ideam_eufo_commission_2020").selfMask()


# Cocoa_bnetd
def nci_ocs2020_prep():
    return (
        ee.Image("BNETD/land_cover/v1/2020")
        .select("classification")
        .eq(9)
        .rename("nCI_Cocoa_bnetd")
    ).selfMask()  # cocoa from national land cover map for Côte d'Ivoire


# nCM - Cameroon
# data from Aurelie Shapiro (FAO) working directly with country experts - info on methods and accuracy assessment to follow


def ncm_treecover_2020_prep():
    return (
        ee.Image("projects/ee-cocoacmr/assets/land_cover/CMR_TNTMMU_2020")
        .select("FNF_2020")
        .eq(1)
        .rename("nCM_Treecover_2020")
        .selfMask()
    )


# ============================================================================
# CONTEXT BANDS (Administrative boundaries and water mask)
# ============================================================================


def g_gaul_admin_code():
    """
    GAUL 2024 Level 1 administrative boundary codes (500m resolution).
    Used for spatial context and administrative aggregation.

    Returns
    -------
    ee.Image
        Image with admin codes renamed to 'admin_code' (as int32)
    """
    admin_image = ee.Image(
        "projects/ee-andyarnellgee/assets/gaul_2024_level_1_code_500m"
    )
    # Cast to int32 to ensure integer GAUL codes, then rename
    return admin_image.rename("admin_code")


def g_water_mask_prep():
    """
    Water mask from JRC/USGS combined dataset.
    Used to identify water bodies for downstream filtering and context.

    Multiplied by pixel area to get water area in hectares.

    Returns
    -------
    ee.Image
        Binary water mask image renamed to In_waterbody (will be multiplied by pixel area)
    """
    from openforis_whisp.parameters.config_runtime import water_flag

    water_mask_image = ee.Image("projects/ee-andyarnellgee/assets/water_mask_jrc_usgs")
    return water_mask_image.selfMask().rename(water_flag)


###Combining datasets


def combine_datasets(
    national_codes=None,
    validate_bands=False,
    include_context_bands=True,
    auto_recovery=False,
):
    """
    Combines datasets into a single multiband image, with fallback if assets are missing.

    Parameters
    ----------
    national_codes : list, optional
        List of ISO2 country codes to include national datasets
    validate_bands : bool, optional
        If True, validates band names with a slow .getInfo() call (default: False)
        Only enable for debugging. Normal operation relies on exception handling.
    include_context_bands : bool, optional
        If True (default), includes context bands (admin_code, water_flag) in the output.
        Set to False when using stats.py implementations that compile datasets differently.
    auto_recovery : bool, optional
        If True (default), automatically enables validate_bands when an error is detected
        during initial assembly. This allows graceful recovery from missing/broken datasets.

    Returns
    -------
    ee.Image
        Combined multiband image with all datasets (and optionally context bands)
    """
    # Step 1: Combine all main dataset images
    all_images = [ee.Image(1).rename(geometry_area_column)]
    for func in list_functions(national_codes=national_codes):
        try:
            all_images.append(func())
        except ee.EEException as e:
            print(f"Error loading image: {e}")

    img_combined = ee.Image.cat(all_images)

    # Step 2: Determine if validation needed
    should_validate = validate_bands
    if auto_recovery and not validate_bands:
        try:
            # Fast error detection: batch check main + context bands in one call
            bands_to_check = [img_combined.bandNames().get(0)]
            if include_context_bands:
                admin_image = g_gaul_admin_code()
                water_mask = g_water_mask_prep()
                bands_to_check.extend(
                    [admin_image.bandNames().get(0), water_mask.bandNames().get(0)]
                )
            ee.List(bands_to_check).getInfo()  # trigger error if any band is invalid
        except ee.EEException as e:
            print(f"Error detected, enabling recovery mode: {str(e)[:80]}...")
            should_validate = True

    # Step 3: Validate and recover if needed
    if should_validate:
        try:
            img_combined.bandNames().getInfo()  # check all bands
        except ee.EEException as e:
            print("Using valid datasets filter due to error in validation")
            valid_imgs = keep_valid_images(
                [func() for func in list_functions(national_codes=national_codes)]
            )
            all_images_retry = [ee.Image(1).rename(geometry_area_column)]
            all_images_retry.extend(valid_imgs)
            img_combined = ee.Image.cat(all_images_retry)

    # Step 4: Multiply main datasets by pixel area
    img_combined = img_combined.multiply(ee.Image.pixelArea())

    # Step 5: Add context bands (admin_code only - water mask is now in prep functions)
    if include_context_bands:
        for band_func, band_name in [
            (g_gaul_admin_code, "admin_code"),
            (g_water_mask_prep, "In_waterbody"),
        ]:
            try:
                band_img = band_func()
                if should_validate:
                    band_img.bandNames().getInfo()
                img_combined = img_combined.addBands(band_img)
            except ee.EEException as e:
                print(f"Warning: Could not add {band_name} band: {e}")

    print("Whisp multiband image compiled")
    return img_combined


######helper functions to check images
# list all functions ending with "_prep" (in the current script)
# def list_functions():
#     # Use the module's globals to get all defined functions
#     current_module = inspect.getmodule(inspect.currentframe())
#     functions = [
#         func
#         for name, func in inspect.getmembers(current_module, inspect.isfunction)
#         if name.endswith("_prep")
#     ]
#     return functions


def list_functions(national_codes=None):
    """
    Returns a list of functions that end with "_prep" and either:
    - Start with "g_" (global/regional products, excluding context bands)
    - Start with any provided national code prefix (nXX_)

    Context band functions (g_gaul_admin_code, g_water_mask_prep) are handled
    separately and excluded from this list to avoid duplication.

    Args:
        national_codes: List of ISO2 country codes (without the 'n' prefix)
    """
    # Use the module's globals to get all defined functions
    current_module = inspect.getmodule(inspect.currentframe())

    # If national_codes is None, default to an empty list
    if national_codes is None:
        national_codes = []

    # Context band functions that are handled separately
    context_functions = {"g_gaul_admin_code", "g_water_mask_prep"}

    # Create prefixes list with proper formatting ('n' + code + '_')
    allowed_prefixes = ["g_"] + [f"n{code.lower()}_" for code in national_codes]

    # Filter functions in a single pass, excluding context band functions
    functions = [
        func
        for name, func in inspect.getmembers(current_module, inspect.isfunction)
        if name.endswith("_prep")
        and any(name.startswith(prefix) for prefix in allowed_prefixes)
        and name not in context_functions
    ]

    return functions


# # IN PROGRESS - expected behaviour
# def filter_by_prefix_list(input_list=None,prefix_list=None):

#     if input_list is None:
#         print ("No function in list")
#     if prefix_list is None:
#         print ("No prefixes listed by which to filter")
#     if input_list is not None:
#         for prefix in prefix_list:
#             if element.startsWith(prefix):
#                 list.


def keep_valid_images(images):
    """Keeps only valid images."""
    valid_images = []
    for img in images:
        try:
            img.getInfo()  # This will raise an exception if the image is invalid
            valid_images.append(img)
        except ee.EEException as e:
            # logger.error(f"Invalid image: {e}")
            print(f"Invalid image: {e}")
    return valid_images


# function to check if an image is valid
def ee_image_checker(image):
    """
    Tests if the input is a valid ee.Image.

    Args:
        image: An ee.Image object.

    Returns:
        bool: True if the input is a valid ee.Image, False otherwise.
    """
    try:
        if ee.Algorithms.ObjectType(image).getInfo() == "Image":
            # Trigger some action on the image to ensure it's a valid image
            image.getInfo()  # This will raise an exception if the image is invalid
            return True
    except ee.EEException as e:
        print(f"Image validation failed with EEException: {e}")
    except Exception as e:
        print(f"Image validation failed with exception: {e}")
    return False


# preparation steps for multiband image with area per pixel values
# function for notebook environment
# user provides custom_images dict and custom_bands_info dict
def combine_custom_bands(custom_images, custom_bands_info):
    """
    Combine custom Earth Engine images into a single multiband image with area conversion.

    Returns
    -------
    ee.Image
        Combined bands converted to area values
    """
    # ... existing validation code ...

    # Step 3: Rename and combine images
    band_names = list(custom_bands_info.keys())

    # Start with first image
    custom_ee_image = custom_images[band_names[0]].rename(band_names[0])

    # Add remaining images if any
    for name in band_names[1:]:
        next_image = custom_images[name].rename(name)
        custom_ee_image = custom_ee_image.addBands(next_image)

    # Convert to area values
    custom_ee_image = custom_ee_image.multiply(ee.Image.pixelArea())

    return custom_ee_image  # Only return the image


# %%



import ee
from google.oauth2 import service_account


def initialize_ee(credentials_path=None, use_high_vol_endpoint=False):
    """Initializes Google Earth Engine using the provided path or defaults to normal if no path is given.
    Args:
    use_high_vol_endpoint: True/False to use high-volume endpoint via opt_url parameter (defaults to False).
    """
    try:
        if not ee.data._initialized:
            print(credentials_path)
            if credentials_path:
                credentials = service_account.Credentials.from_service_account_file(
                    credentials_path,
                    scopes=["https://www.googleapis.com/auth/earthengine"],
                )
                if use_high_vol_endpoint == False:
                    ee.Initialize(credentials)
                else:
                    ee.Initialize(
                        credentials,
                        opt_url="https://earthengine-highvolume.googleapis.com",
                    )
                print("EE initialized with credentials from:", credentials_path)
            else:
                if use_high_vol_endpoint == False:
                    ee.Initialize()
                else:
                    ee.Initialize(
                        opt_url="https://earthengine-highvolume.googleapis.com"
                    )
                print("EE initialized with default credentials.")
    except Exception as e:
        print("Error initializing EE:", e)


# Default to normal initialize if nobody calls whisp.initialize_ee.
try:
    if not ee.data._initialized:
        ee.Initialize()
        print("EE auto-initialized with default credentials.")
except Exception as e:
    print("Error in default EE initialization:", e)

from openforis_whisp.datasets import combine_datasets, combine_custom_bands

from openforis_whisp.stats import (
    whisp_stats_ee_to_ee,
    whisp_stats_ee_to_df,
    whisp_stats_geojson_to_df,
    whisp_stats_geojson_to_ee,
    whisp_stats_geojson_to_geojson,
    whisp_stats_ee_to_drive,
    whisp_stats_geojson_to_drive,
    whisp_formatted_stats_ee_to_df,
    whisp_formatted_stats_ee_to_geojson,
    whisp_formatted_stats_geojson_to_df,
    whisp_formatted_stats_geojson_to_geojson,
    set_point_geometry_area_to_zero,
    reformat_geometry_type,
    convert_iso3_to_iso2,
)

from openforis_whisp.advanced_stats import (
    whisp_formatted_stats_geojson_to_df_fast,
)

# temporary parameters to be removed once isio3 to iso2 conversion server side is implemented
from openforis_whisp.parameters.config_runtime import (
    iso3_country_column,
    iso2_country_column,
)

from openforis_whisp.reformat import (
    validate_dataframe_using_lookups,
    validate_dataframe_using_lookups_flexible,
    validate_dataframe,
    create_schema_from_dataframe,
    load_schema_if_any_file_changed,
    format_stats_dataframe,
)

from openforis_whisp.data_conversion import (
    convert_ee_to_df,
    convert_geojson_to_ee,
    convert_df_to_geojson,
    convert_csv_to_geojson,
    convert_ee_to_geojson,
)

from openforis_whisp.risk import whisp_risk, detect_unit_type

from openforis_whisp.utils import (
    get_example_data_path,
    generate_test_polygons,  # to be deprecated
    generate_random_features,
    generate_random_points,
    generate_random_polygons,
)

from openforis_whisp.data_checks import (
    analyze_geojson,
    check_geojson_limits,
    screen_geojson,  # Backward compatibility alias
    suggest_processing_mode,
    validate_geojson_constraints,  # Backward compatibility alias
)



import logging
import sys

BASE_MSG_FORMAT = (
    "[%(filename)s | %(funcName)s() | l.%(lineno)s] %(levelname)s: %(message)s"
)


class StdoutLogger:
    def __init__(self, name: str, msg_format: str = BASE_MSG_FORMAT) -> None:
        # Create handler that auto-flushes for Colab/notebook visibility
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(logging.Formatter(msg_format))
        handler.setLevel(logging.DEBUG)

        # Override emit to force flush after each message
        original_emit = handler.emit

        def emit_with_flush(record):
            original_emit(record)
            sys.stdout.flush()

        handler.emit = emit_with_flush

        self.handler = handler
        self.logger = logging.getLogger(name)
        self.logger.addHandler(self.handler)
        self.logger.propagate = False

    # Add missing methods that delegate to the internal logger
    def debug(self, message):
        self.logger.debug(message)

    def info(self, message):
        self.logger.info(message)

    def warning(self, message):
        self.logger.warning(message)

    def error(self, message):
        self.logger.error(message)

    def critical(self, message):
        self.logger.critical(message)

    def setLevel(self, level):
        self.logger.setLevel(level)

    @property
    def level(self):
        """Return the logger's effective level."""
        return self.logger.level

    def hasHandlers(self):
        """Check if the logger has any handlers."""
        return self.logger.hasHandlers()

    def addHandler(self, handler):
        """Add a handler to the logger."""
        self.logger.addHandler(handler)


class FileLogger:
    def __init__(
        self,
        log_filepath: str,
        msg_format: str = BASE_MSG_FORMAT,
        log_to_stdout: bool = True,
    ) -> None:
        self.handler = logging.FileHandler(log_filepath)
        self.handler.setFormatter(logging.Formatter(msg_format))
        self.handler.setLevel(logging.DEBUG)
        self.logger = logging.getLogger(f"{__name__}.file_logger_{log_filepath}")
        self.logger.addHandler(self.handler)
        self.logger.propagate = False

        if log_to_stdout:
            self.stdout_handler = logging.StreamHandler(sys.stdout)
            self.stdout_handler.setFormatter(logging.Formatter(msg_format))
            self.stdout_handler.setLevel(logging.DEBUG)
            self.logger.addHandler(self.stdout_handler)

    # Add missing methods for FileLogger too
    def debug(self, message):
        self.logger.debug(message)

    def info(self, message):
        self.logger.info(message)

    def warning(self, message):
        self.logger.warning(message)

    def error(self, message):
        self.logger.error(message)

    def critical(self, message):
        self.logger.critical(message)

    def setLevel(self, level):
        self.logger.setLevel(level)

    @property
    def level(self):
        """Return the logger's effective level."""
        return self.logger.level

    def hasHandlers(self):
        """Check if the logger has any handlers."""
        return self.logger.hasHandlers()

    def addHandler(self, handler):
        """Add a handler to the logger."""
        self.logger.addHandler(handler)
